import ast
from collections.abc import Iterator, Sequence
from contextlib import AbstractContextManager, ExitStack, contextmanager
from typing import Any, Final, TypeGuard, TypeVar

import hugr
import hugr.std.collections.array
import hugr.std.float
import hugr.std.int
import hugr.std.logic
import hugr.std.prelude
from hugr import Wire, ops
from hugr import tys as ht
from hugr import val as hv
from hugr.build import function as hf
from hugr.build.cond_loop import Conditional
from hugr.build.dfg import DP, DfBase
from typing_extensions import assert_never

from guppylang_internals.ast_util import AstNode, AstVisitor, get_type
from guppylang_internals.cfg.builder import tmp_vars
from guppylang_internals.checker.core import Variable, contains_subscript
from guppylang_internals.checker.errors.generic import UnsupportedError
from guppylang_internals.compiler.core import (
    DEBUG_EXTENSION,
    RESULT_EXTENSION,
    CompilerBase,
    CompilerContext,
    DFContainer,
    GlobalConstId,
)
from guppylang_internals.compiler.hugr_extension import PartialOp
from guppylang_internals.definition.custom import CustomFunctionDef
from guppylang_internals.definition.value import (
    CallableDef,
    CallReturnWires,
    CompiledCallableDef,
    CompiledValueDef,
)
from guppylang_internals.engine import ENGINE
from guppylang_internals.error import GuppyError, InternalGuppyError
from guppylang_internals.nodes import (
    BarrierExpr,
    DesugaredArrayComp,
    DesugaredGenerator,
    DesugaredListComp,
    ExitKind,
    FieldAccessAndDrop,
    GenericParamValue,
    GlobalCall,
    GlobalName,
    LocalCall,
    PanicExpr,
    PartialApply,
    PlaceNode,
    ResultExpr,
    StateResultExpr,
    SubscriptAccessAndDrop,
    TensorCall,
    TupleAccessAndDrop,
    TypeApply,
)
from guppylang_internals.std._internal.compiler.arithmetic import (
    UnsignedIntVal,
    convert_ifromusize,
)
from guppylang_internals.std._internal.compiler.array import (
    array_convert_from_std_array,
    array_convert_to_std_array,
    array_map,
    array_new,
    array_repeat,
    standard_array_type,
    unpack_array,
)
from guppylang_internals.std._internal.compiler.list import (
    list_new,
)
from guppylang_internals.std._internal.compiler.prelude import (
    build_error,
    build_panic,
    build_unwrap,
    panic,
)
from guppylang_internals.std._internal.compiler.tket_bool import (
    OpaqueBool,
    OpaqueBoolVal,
    make_opaque,
    not_op,
    read_bool,
)
from guppylang_internals.tys.arg import ConstArg
from guppylang_internals.tys.builtin import (
    bool_type,
    get_element_type,
    int_type,
    is_bool_type,
    is_frozenarray_type,
)
from guppylang_internals.tys.const import ConstValue
from guppylang_internals.tys.subst import Inst
from guppylang_internals.tys.ty import (
    BoundTypeVar,
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    OpaqueType,
    TupleType,
    Type,
    type_to_row,
)


class ExprCompiler(CompilerBase, AstVisitor[Wire]):
    """A compiler from guppylang expressions to Hugr."""

    dfg: DFContainer

    def compile(self, expr: ast.expr, dfg: DFContainer) -> Wire:
        """Compiles an expression and returns a single wire holding the output value."""
        self.dfg = dfg
        return self.visit(expr)

    def compile_row(self, expr: ast.expr, dfg: DFContainer) -> list[Wire]:
        """Compiles a row expression and returns a list of wires, one for each value in
        the row.

        On Python-level, we treat tuples like rows on top-level. However, nested tuples
        are treated like regular Guppy tuples.
        """
        return [self.compile(e, dfg) for e in expr_to_row(expr)]

    @property
    def builder(self) -> DfBase[ops.DfParentOp]:
        """The current Hugr dataflow graph builder."""
        return self.dfg.builder

    @contextmanager
    def _new_dfcontainer(
        self, inputs: list[PlaceNode], builder: DfBase[DP]
    ) -> Iterator[None]:
        """Context manager to build a graph inside a new `DFContainer`.

        Automatically updates `self.dfg` and makes the inputs available.
        """
        old = self.dfg
        # Check that the input names are unique
        assert len({inp.place.id for inp in inputs}) == len(
            inputs
        ), "Inputs are not unique"
        self.dfg = DFContainer(builder, self.ctx, self.dfg.locals.copy())
        hugr_input = builder.input_node
        for input_node, wire in zip(inputs, hugr_input, strict=True):
            self.dfg[input_node.place] = wire

        yield

        self.dfg = old

    @contextmanager
    def _new_loop(
        self,
        just_inputs_vars: list[PlaceNode],
        loop_vars: list[PlaceNode],
        break_predicate: PlaceNode,
    ) -> Iterator[None]:
        """Context manager to build a graph inside a new `TailLoop` node.

        Automatically adds the `Output` node to the loop body once the context manager
        exits.
        """
        just_inputs = [self.visit(name) for name in just_inputs_vars]
        loop_inputs = [self.visit(name) for name in loop_vars]
        loop = self.builder.add_tail_loop(just_inputs, loop_inputs)
        with self._new_dfcontainer(just_inputs_vars + loop_vars, loop):
            yield
            # Output the branch predicate and the inputs for the next iteration. Note
            # that we have to do fresh calls to `self.visit` here since we're in a new
            # context
            do_break = self.visit(break_predicate)
            loop.set_loop_outputs(do_break, *(self.visit(name) for name in loop_vars))
        # Update the DFG with the outputs from the loop
        for node, wire in zip(loop_vars, loop, strict=True):
            self.dfg[node.place] = wire

    @contextmanager
    def _new_case(
        self,
        inputs: list[PlaceNode],
        outputs: list[PlaceNode],
        conditional: Conditional,
        case_id: int,
    ) -> Iterator[None]:
        """Context manager to build a graph inside a new `Case` node.

        Automatically adds the `Output` node once the context manager exits.
        """
        # TODO: `Case` is `_DfgBase`, but not `Dfg`?
        case = conditional.add_case(case_id)
        with self._new_dfcontainer(inputs, case):
            yield
            case.set_outputs(*(self.visit(name) for name in outputs))

    def _if_else(
        self,
        cond: ast.expr,
        inputs: list[PlaceNode],
        outputs: list[PlaceNode],
        only_true_inputs: list[PlaceNode] | None = None,
        only_false_inputs: list[PlaceNode] | None = None,
    ) -> tuple[AbstractContextManager[None], AbstractContextManager[None]]:
        """Builds a `Conditional`, returning context managers to build the `True` and
        `False` branch.
        """
        cond_wire = self.visit(cond)
        cond_ty = self.builder.hugr.port_type(cond_wire.out_port())
        if cond_ty == OpaqueBool:
            cond_wire = self.builder.add_op(read_bool(), cond_wire)
        conditional = self.builder.add_conditional(
            cond_wire, *(self.visit(inp) for inp in inputs)
        )
        only_true_inputs_ = only_true_inputs or []
        only_false_inputs_ = only_false_inputs or []

        @contextmanager
        def true_case() -> Iterator[None]:
            with self._new_case(only_true_inputs_ + inputs, outputs, conditional, 1):
                yield

        @contextmanager
        def false_case() -> Iterator[None]:
            with self._new_case(only_false_inputs_ + inputs, outputs, conditional, 0):
                yield
            # Update the DFG with the outputs from the Conditional node
            for node, wire in zip(outputs, conditional, strict=True):
                self.dfg[node.place] = wire

        return true_case(), false_case()

    @contextmanager
    def _if_true(self, cond: ast.expr, inputs: list[PlaceNode]) -> Iterator[None]:
        """Context manager to build a graph inside the `true` case of a `Conditional`

        In the `false` case, the inputs are outputted as is.
        """
        true_case, false_case = self._if_else(cond, inputs, inputs)
        with false_case:
            # If the condition is false, output the inputs as is
            pass
        with true_case:
            # If the condition is true, we enter the `with` block
            yield

    def visit_Constant(self, node: ast.Constant) -> Wire:
        if value := python_value_to_hugr(node.value, get_type(node), self.ctx):
            return self.builder.load(value)
        raise InternalGuppyError("Unsupported constant expression in compiler")

    def visit_PlaceNode(self, node: PlaceNode) -> Wire:
        if subscript := contains_subscript(node.place):
            if subscript.item not in self.dfg:
                self.dfg[subscript.item] = self.visit(subscript.item_expr)
            self.dfg[subscript] = self.visit(subscript.getitem_call)
        return self.dfg[node.place]

    def visit_GlobalName(self, node: GlobalName) -> Wire:
        defn = ENGINE.get_checked(node.def_id)
        if isinstance(defn, CallableDef) and defn.ty.parametrized:
            # TODO: This should be caught during checking
            err = UnsupportedError(
                node, "Polymorphic functions as dynamic higher-order values"
            )
            raise GuppyError(err)

        defn, [] = self.ctx.build_compiled_def(node.def_id, type_args=[])
        assert isinstance(defn, CompiledValueDef)
        return defn.load(self.dfg, self.ctx, node)

    def visit_GenericParamValue(self, node: GenericParamValue) -> Wire:
        match node.param.ty:
            case NumericType(NumericType.Kind.Nat):
                # Generic nat parameters are encoded using Hugr bounded nat parameters,
                # so they are not monomorphized when compiling to Hugr
                arg = node.param.to_bound().to_hugr(self.ctx)
                load_nat = hugr.std.PRELUDE.get_op("load_nat").instantiate(
                    [arg], ht.FunctionType([], [ht.USize()])
                )
                usize = self.builder.add_op(load_nat)
                return self.builder.add_op(convert_ifromusize(), usize)
            case ty:
                # Look up monomorphization
                assert self.ctx.current_mono_args is not None
                match self.ctx.current_mono_args[node.param.idx]:
                    case ConstArg(const=ConstValue(value=v)):
                        val = python_value_to_hugr(v, ty, self.ctx)
                        assert val is not None
                        return self.builder.load(val)
                    case _:
                        raise InternalGuppyError("Monomorphized const is not a value")

    def visit_Name(self, node: ast.Name) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_Tuple(self, node: ast.Tuple) -> Wire:
        elems = [self.visit(e) for e in node.elts]
        types = [get_type(e) for e in node.elts]
        return self._pack_tuple(elems, types)

    def visit_List(self, node: ast.List) -> Wire:
        # Note that this is a list literal (i.e. `[e1, e2, ...]`), not a comprehension
        inputs = [self.visit(e) for e in node.elts]
        list_ty = get_type(node)
        elem_ty = get_element_type(list_ty)
        return list_new(self.builder, elem_ty.to_hugr(self.ctx), inputs)

    def _unpack_tuple(self, wire: Wire, types: Sequence[Type]) -> Sequence[Wire]:
        """Add a tuple unpack operation to the graph"""
        types = [t.to_hugr(self.ctx) for t in types]
        return list(self.builder.add_op(ops.UnpackTuple(types), wire))

    def _pack_tuple(self, wires: Sequence[Wire], types: Sequence[Type]) -> Wire:
        """Add a tuple pack operation to the graph"""
        types = [t.to_hugr(self.ctx) for t in types]
        return self.builder.add_op(ops.MakeTuple(types), *wires)

    def _pack_returns(self, returns: Sequence[Wire], return_ty: Type) -> Wire:
        """Groups function return values into a tuple"""
        if isinstance(return_ty, TupleType | NoneType) and not return_ty.preserve:
            types = type_to_row(return_ty)
            assert len(returns) == len(types)
            return self._pack_tuple(returns, types)
        assert (
            len(returns) == 1
        ), f"Expected a single return value. Got {returns}. return type {return_ty}"
        return returns[0]

    def _update_inout_ports(
        self,
        args: list[ast.expr],
        inout_ports: Iterator[Wire],
        func_ty: FunctionType,
    ) -> None:
        """Helper method that updates the ports for borrowed arguments after a call."""
        for inp, arg in zip(func_ty.inputs, args, strict=True):
            if InputFlags.Inout in inp.flags:
                # Linearity checker ensures that borrowed arguments that are not places
                # can be safely dropped after the call returns
                if not isinstance(arg, PlaceNode):
                    next(inout_ports)
                    continue
                self.dfg[arg.place] = next(inout_ports)
                # Places involving subscripts need to generate code for the appropriate
                # `__setitem__` call. Nested subscripts are handled automatically since
                # `arg.place.parent` occurs as an arg of this call, so will also
                # be recursively reassigned.
                if subscript := contains_subscript(arg.place):
                    assert subscript.setitem_call is not None
                    # Need to assign __setitem__ value before compiling call.
                    # Note that the assignment to `self.dfg[arg.place]` also updated
                    # `self.dfg[subscript]` so that it now contains the value we want
                    # to write back into the subscript.
                    self.dfg[subscript.setitem_call.value_var] = self.dfg[subscript]
                    self.visit(subscript.setitem_call.call)
        assert next(inout_ports, None) is None, "Too many inout return ports"

    def visit_LocalCall(self, node: LocalCall) -> Wire:
        func = self.visit(node.func)
        func_ty = get_type(node.func)
        assert isinstance(func_ty, FunctionType)
        num_returns = len(type_to_row(func_ty.output))

        args = self._compile_call_args(node.args, func_ty)
        call = self.builder.add_op(
            ops.CallIndirect(func_ty.to_hugr(self.ctx)), func, *args
        )
        regular_returns = list(call[:num_returns])
        inout_returns = call[num_returns:]
        self._update_inout_ports(node.args, inout_returns, func_ty)
        return self._pack_returns(regular_returns, func_ty.output)

    def visit_TensorCall(self, node: TensorCall) -> Wire:
        functions: Wire = self.visit(node.func)
        function_types = get_type(node.func)
        assert isinstance(function_types, TupleType)

        rets: list[Wire] = []
        remaining_args = node.args
        for func, func_ty in zip(
            self._unpack_tuple(functions, function_types.element_types),
            function_types.element_types,
            strict=True,
        ):
            outs, remaining_args = self._compile_tensor_with_leftovers(
                func, func_ty, remaining_args
            )
            rets.extend(outs)
        assert (
            remaining_args == []
        ), "Not all function arguments were consumed after a tensor call"
        return self._pack_returns(rets, node.tensor_ty.output)

    def _compile_tensor_with_leftovers(
        self, func: Wire, func_ty: Type, args: list[ast.expr]
    ) -> tuple[
        list[Wire],  # Compiled outputs
        list[ast.expr],  # Leftover args
    ]:
        """Compiles a function call, consuming as many arguments as needed, and
        returning the unused ones.
        """
        if isinstance(func_ty, TupleType):
            remaining_args = args
            all_outs = []
            for elem, ty in zip(
                self._unpack_tuple(func, func_ty.element_types),
                func_ty.element_types,
                strict=True,
            ):
                outs, remaining_args = self._compile_tensor_with_leftovers(
                    elem, ty, remaining_args
                )
                all_outs.extend(outs)
            return all_outs, remaining_args

        elif isinstance(func_ty, FunctionType):
            input_len = len(func_ty.inputs)
            num_returns = len(type_to_row(func_ty.output))
            consumed_args, other_args = args[0:input_len], args[input_len:]
            consumed_wires = self._compile_call_args(consumed_args, func_ty)
            call = self.builder.add_op(
                ops.CallIndirect(func_ty.to_hugr(self.ctx)), func, *consumed_wires
            )
            regular_returns: list[Wire] = list(call[:num_returns])
            inout_returns = call[num_returns:]
            self._update_inout_ports(consumed_args, inout_returns, func_ty)
            return regular_returns, other_args
        else:
            raise InternalGuppyError("Tensor element wasn't function or tuple")

    def visit_GlobalCall(self, node: GlobalCall) -> Wire:
        func, rem_args = self.ctx.build_compiled_def(node.def_id, node.type_args)
        assert isinstance(func, CompiledCallableDef)

        if isinstance(func, CustomFunctionDef) and not func.has_signature:
            func_ty = FunctionType(
                [FuncInput(get_type(arg), InputFlags.NoFlags) for arg in node.args],
                get_type(node),
            )
        else:
            func_ty = func.ty.instantiate(rem_args)

        args = self._compile_call_args(node.args, func_ty)
        rets = func.compile_call(args, rem_args, self.dfg, self.ctx, node)
        self._update_inout_ports(node.args, iter(rets.inout_returns), func_ty)
        return self._pack_returns(rets.regular_returns, func_ty.output)

    def _compile_call_args(
        self, args: list[ast.expr], func_ty: FunctionType
    ) -> list[Wire]:
        """Helper function to compile arguments for function calls.

        Takes care of filtering out comptime arguments that are provided via generic
        args or monomorphization instead of wires.
        """
        return [
            self.visit(arg)
            for arg, inp in zip(args, func_ty.inputs, strict=True)
            # Don't compile comptime args since we already include them as type args
            if InputFlags.Comptime not in inp.flags
        ]

    def visit_Call(self, node: ast.Call) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_PartialApply(self, node: PartialApply) -> Wire:
        func_ty = get_type(node.func)
        assert isinstance(func_ty, FunctionType)
        op = PartialOp.from_closure(
            func_ty.to_hugr(self.ctx),
            [get_type(arg).to_hugr(self.ctx) for arg in node.args],
        )
        return self.builder.add_op(
            op, self.visit(node.func), *(self.visit(arg) for arg in node.args)
        )

    def visit_TypeApply(self, node: TypeApply) -> Wire:
        # For now, we can only TypeApply global FunctionDefs/Decls.
        if not isinstance(node.value, GlobalName):
            raise InternalGuppyError("Dynamic TypeApply not supported yet!")
        defn, rem_args = self.ctx.build_compiled_def(node.value.def_id, node.inst)
        assert isinstance(defn, CompiledCallableDef)

        # We have to be very careful here: If we instantiate `foo: forall T. T -> T`
        # with a tuple type `tuple[A, B]`, we get the type `tuple[A, B] -> tuple[A, B]`.
        # Normally, this would be represented in Hugr as a function with two output
        # ports types A and B. However, when TypeApplying `foo`, we actually get a
        # function with a single output port typed `tuple[A, B]`.
        # TODO: We would need to do manual monomorphisation in that case to obtain a
        #  function that returns two ports as expected
        if instantiation_needs_unpacking(defn.ty, node.inst):
            err = UnsupportedError(
                node, "Generic function instantiations returning rows"
            )
            raise GuppyError(err)

        return defn.load_with_args(rem_args, self.dfg, self.ctx, node)

    def visit_UnaryOp(self, node: ast.UnaryOp) -> Wire:
        # The only case that is not desugared by the type checker is the `not` operation
        # since it is not implemented via a dunder method
        if isinstance(node.op, ast.Not):
            arg = self.visit(node.operand)
            return self.builder.add_op(not_op(), arg)

        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_FieldAccessAndDrop(self, node: FieldAccessAndDrop) -> Wire:
        struct_port = self.visit(node.value)
        field_idx = node.struct_ty.fields.index(node.field)
        return self._unpack_tuple(struct_port, [f.ty for f in node.struct_ty.fields])[
            field_idx
        ]

    def visit_SubscriptAccessAndDrop(self, node: SubscriptAccessAndDrop) -> Wire:
        self.dfg[node.item] = self.visit(node.item_expr)
        return self.visit(node.getitem_expr)

    def visit_TupleAccessAndDrop(self, node: TupleAccessAndDrop) -> Wire:
        tuple_port = self.visit(node.value)
        return self._unpack_tuple(tuple_port, node.tuple_ty.element_types)[node.index]

    def visit_ResultExpr(self, node: ResultExpr) -> Wire:
        value_wire = self.visit(node.value)
        base_ty = node.base_ty.to_hugr(self.ctx)
        extra_args: list[ht.TypeArg] = []
        if isinstance(node.base_ty, NumericType):
            match node.base_ty.kind:
                case NumericType.Kind.Nat:
                    base_name = "uint"
                    extra_args = [ht.BoundedNatArg(n=NumericType.INT_WIDTH)]
                case NumericType.Kind.Int:
                    base_name = "int"
                    extra_args = [ht.BoundedNatArg(n=NumericType.INT_WIDTH)]
                case NumericType.Kind.Float:
                    base_name = "f64"
                case kind:
                    assert_never(kind)
        else:
            # The only other valid base type is bool
            assert is_bool_type(node.base_ty)
            base_name = "bool"
        if node.array_len is not None:
            op_name = f"result_array_{base_name}"
            size_arg = node.array_len.to_arg().to_hugr(self.ctx)
            extra_args = [size_arg, *extra_args]
            # Remove the option wrapping in the array
            unwrap = array_unwrap_elem(self.ctx)
            unwrap = self.builder.load_function(
                unwrap,
                instantiation=ht.FunctionType([ht.Option(base_ty)], [base_ty]),
                type_args=[ht.TypeTypeArg(base_ty)],
            )
            map_op = array_map(ht.Option(base_ty), size_arg, base_ty)
            value_wire = self.builder.add_op(map_op, value_wire, unwrap)
            if is_bool_type(node.base_ty):
                # We need to coerce a read on all the array elements if they are bools.
                array_read = array_read_bool(self.ctx)
                array_read = self.builder.load_function(array_read)
                map_op = array_map(OpaqueBool, size_arg, ht.Bool)
                value_wire = self.builder.add_op(map_op, value_wire, array_read)
                base_ty = ht.Bool
            # Turn `value_array` into regular linear `array`
            value_wire = self.builder.add_op(
                array_convert_to_std_array(base_ty, size_arg), value_wire
            )
            hugr_ty: ht.Type = hugr.std.collections.array.Array(base_ty, size_arg)
        else:
            if is_bool_type(node.base_ty):
                base_ty = ht.Bool
                value_wire = self.builder.add_op(read_bool(), value_wire)
            op_name = f"result_{base_name}"
            hugr_ty = base_ty

        sig = ht.FunctionType(input=[hugr_ty], output=[])
        args = [ht.StringArg(node.tag), *extra_args]
        op = ops.ExtOp(RESULT_EXTENSION.get_op(op_name), signature=sig, args=args)

        self.builder.add_op(op, value_wire)
        return self._pack_returns([], NoneType())

    def visit_PanicExpr(self, node: PanicExpr) -> Wire:
        err = build_error(self.builder, node.signal, node.msg)
        in_tys = [get_type(e).to_hugr(self.ctx) for e in node.values]
        out_tys = [ty.to_hugr(self.ctx) for ty in type_to_row(get_type(node))]
        args = [self.visit(e) for e in node.values]
        match node.kind:
            case ExitKind.Panic:
                h_node = build_panic(self.builder, in_tys, out_tys, err, *args)
            case ExitKind.ExitShot:
                op = panic(in_tys, out_tys, ExitKind.ExitShot)
                h_node = self.builder.add_op(op, err, *args)
        return self._pack_returns(list(h_node.outputs()), get_type(node))

    def visit_BarrierExpr(self, node: BarrierExpr) -> Wire:
        hugr_tys = [get_type(e).to_hugr(self.ctx) for e in node.args]
        op = hugr.std.prelude.PRELUDE_EXTENSION.get_op("Barrier").instantiate(
            [ht.ListArg([ht.TypeTypeArg(ty) for ty in hugr_tys])],
            ht.FunctionType.endo(hugr_tys),
        )

        barrier_n = self.builder.add_op(op, *(self.visit(e) for e in node.args))

        self._update_inout_ports(node.args, iter(barrier_n), node.func_ty)
        return self._pack_returns([], NoneType())

    def visit_StateResultExpr(self, node: StateResultExpr) -> Wire:
        num_qubits_arg = (
            node.array_len.to_arg().to_hugr(self.ctx)
            if node.array_len
            else ht.BoundedNatArg(len(node.args) - 1)
        )
        args = [ht.StringArg(node.tag), num_qubits_arg]
        sig = ht.FunctionType(
            [standard_array_type(ht.Qubit, num_qubits_arg)],
            [standard_array_type(ht.Qubit, num_qubits_arg)],
        )

        op = ops.ExtOp(DEBUG_EXTENSION.get_op("StateResult"), signature=sig, args=args)

        if not node.array_len:
            # If the input is a sequence of qubits, we pack them into an array.
            qubits_in = [self.visit(e) for e in node.args[1:]]
            qubit_arr_in = self.builder.add_op(
                array_new(ht.Qubit, len(node.args) - 1), *qubits_in
            )
            # Turn into standard array from value array.
            qubit_arr_in = self.builder.add_op(
                array_convert_to_std_array(ht.Qubit, num_qubits_arg), qubit_arr_in
            )

            qubit_arr_out = self.builder.add_op(op, qubit_arr_in)

            qubit_arr_out = self.builder.add_op(
                array_convert_from_std_array(ht.Qubit, num_qubits_arg), qubit_arr_out
            )
            qubits_out = unpack_array(self.builder, qubit_arr_out)
        else:
            # If the input is an array of qubits, we need to unwrap the elements first,
            # and then convert to a value array and back.
            qubits_in = [self.visit(node.args[1])]
            qubits_out = [
                apply_array_op_with_conversions(
                    self.ctx, self.builder, op, ht.Qubit, num_qubits_arg, qubits_in[0]
                )
            ]

        self._update_inout_ports(node.args, iter(qubits_out), node.func_ty)
        return self._pack_returns([], NoneType())

    def visit_DesugaredListComp(self, node: DesugaredListComp) -> Wire:
        # Make up a name for the list under construction and bind it to an empty list
        list_ty = get_type(node)
        assert isinstance(list_ty, OpaqueType)
        elem_ty = get_element_type(list_ty)
        list_place = Variable(next(tmp_vars), list_ty, node)
        self.dfg[list_place] = list_new(self.builder, elem_ty.to_hugr(self.ctx), [])
        with self._build_generators(node.generators, [list_place]):
            elt_port = self.visit(node.elt)
            list_port = self.dfg[list_place]
            [], [self.dfg[list_place]] = self._build_method_call(
                list_ty, "append", node, [list_port, elt_port], list_ty.args
            )
        return self.dfg[list_place]

    def visit_DesugaredArrayComp(self, node: DesugaredArrayComp) -> Wire:
        # Allocate an uninitialised array of the desired size and a counter variable
        array_ty = get_type(node)
        assert isinstance(array_ty, OpaqueType)
        array_var = Variable(next(tmp_vars), array_ty, node)
        count_var = Variable(next(tmp_vars), int_type(), node)
        # See https://github.com/CQCL/guppylang/issues/629
        hugr_elt_ty = ht.Option(node.elt_ty.to_hugr(self.ctx))
        # Initialise array with `None`s
        make_none = array_comprehension_init_func(self.ctx)
        make_none = self.builder.load_function(
            make_none,
            instantiation=ht.FunctionType([], [hugr_elt_ty]),
            type_args=[ht.TypeTypeArg(node.elt_ty.to_hugr(self.ctx))],
        )
        self.dfg[array_var] = self.builder.add_op(
            array_repeat(hugr_elt_ty, node.length.to_arg().to_hugr(self.ctx)), make_none
        )
        self.dfg[count_var] = self.builder.load(
            hugr.std.int.IntVal(0, width=NumericType.INT_WIDTH)
        )
        with self._build_generators([node.generator], [array_var, count_var]):
            elt = self.visit(node.elt)
            array, count = self.dfg[array_var], self.dfg[count_var]
            [], [self.dfg[array_var]] = self._build_method_call(
                array_ty, "__setitem__", node, [array, count, elt], array_ty.args
            )
            # Update `count += 1`
            one = self.builder.load(hugr.std.int.IntVal(1, width=NumericType.INT_WIDTH))
            [self.dfg[count_var]], [] = self._build_method_call(
                int_type(), "__add__", node, [count, one], []
            )
        return self.dfg[array_var]

    def _build_method_call(
        self, ty: Type, method: str, node: AstNode, args: list[Wire], type_args: Inst
    ) -> CallReturnWires:
        func_and_targs = self.ctx.build_compiled_instance_func(ty, method, type_args)
        assert func_and_targs is not None
        func, rem_args = func_and_targs
        return func.compile_call(args, rem_args, self.dfg, self.ctx, node)

    @contextmanager
    def _build_generators(
        self, gens: list[DesugaredGenerator], loop_vars: list[Variable]
    ) -> Iterator[None]:
        """Context manager to build and enter the `TailLoop`s for a list of generators.

        The provided `loop_vars` will be threaded through and will be available inside
        the loops.
        """
        from guppylang_internals.compiler.stmt_compiler import StmtCompiler

        compiler = StmtCompiler(self.ctx)
        with ExitStack() as stack:
            for gen in gens:
                # Build the generator
                compiler.compile_stmts([gen.iter_assign], self.dfg)
                assert isinstance(gen.iter, PlaceNode)
                iter_ty = get_type(gen.iter)
                inputs = [PlaceNode(place=var) for var in loop_vars]
                inputs += [PlaceNode(place=place) for place in gen.used_outer_places]
                # Enter a new tail loop. Note that the iterator is a `just_input`, so
                # will not be outputted by the loop
                break_pred = PlaceNode(Variable(next(tmp_vars), bool_type(), gen.iter))
                stack.enter_context(self._new_loop([gen.iter], inputs, break_pred))
                # Enter a conditional checking if we have a next element
                next_ty = TupleType([get_type(gen.target), iter_ty])
                next_var = PlaceNode(Variable(next(tmp_vars), next_ty, gen.iter))
                hasnext_case, stop_case = self._if_else(
                    gen.next_call,
                    inputs,
                    only_true_inputs=[next_var],
                    outputs=[break_pred, *inputs],
                )
                # In the "no" case, we set the break predicate to true
                break_pred_hugr_ty = ht.Either([iter_ty.to_hugr(self.ctx)], [])
                with stop_case:
                    self.dfg[break_pred.place] = self.dfg.builder.add_op(
                        ops.Tag(1, break_pred_hugr_ty)
                    )
                # Otherwise, we continue, set the break predicate to false, and insert
                # the iterator for the next loop iteration
                stack.enter_context(hasnext_case)
                next_wire = self.dfg[next_var.place]
                elt, it = self.dfg.builder.add_op(ops.UnpackTuple(), next_wire)
                compiler.dfg = self.dfg
                compiler._assign(gen.target, elt)
                self.dfg[break_pred.place] = self.dfg.builder.add_op(
                    ops.Tag(0, break_pred_hugr_ty), it
                )
                # Enter nested conditionals for each if guard on the generator
                for if_expr in gen.ifs:
                    stack.enter_context(self._if_true(if_expr, [break_pred, *inputs]))
            # Yield control to the caller to build inside the loop
            yield

    def visit_BinOp(self, node: ast.BinOp) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")

    def visit_Compare(self, node: ast.Compare) -> Wire:
        raise InternalGuppyError("Node should have been removed during type checking.")


def expr_to_row(expr: ast.expr) -> list[ast.expr]:
    """Turns an expression into a row expressions by unpacking top-level tuples."""
    return expr.elts if isinstance(expr, ast.Tuple) else [expr]


def instantiation_needs_unpacking(func_ty: FunctionType, inst: Inst) -> bool:
    """Checks if instantiating a polymorphic makes it return a row."""
    if isinstance(func_ty.output, BoundTypeVar):
        return_ty = inst[func_ty.output.idx]
        return isinstance(return_ty, TupleType | NoneType)
    return False


def python_value_to_hugr(v: Any, exp_ty: Type, ctx: CompilerContext) -> hv.Value | None:
    """Turns a Python value into a Hugr value.

    Returns None if the Python value cannot be represented in Guppy.
    """
    match v:
        case bool():
            return OpaqueBoolVal(v)
        case str():
            return hugr.std.prelude.StringVal(v)
        case int():
            assert isinstance(exp_ty, NumericType)
            match exp_ty.kind:
                case NumericType.Kind.Nat:
                    return UnsignedIntVal(v, width=NumericType.INT_WIDTH)
                case NumericType.Kind.Int:
                    return hugr.std.int.IntVal(v, width=NumericType.INT_WIDTH)
                case _:
                    raise InternalGuppyError("Unexpected numeric type")
        case float():
            return hugr.std.float.FloatVal(v)
        case tuple(elts):
            assert isinstance(exp_ty, TupleType)
            vs = [
                python_value_to_hugr(elt, ty, ctx)
                for elt, ty in zip(elts, exp_ty.element_types, strict=True)
            ]
            if doesnt_contain_none(vs):
                return hv.Tuple(*vs)
        case list(elts):
            assert is_frozenarray_type(exp_ty)
            elem_ty = get_element_type(exp_ty)
            vs = [python_value_to_hugr(elt, elem_ty, ctx) for elt in elts]
            if doesnt_contain_none(vs):
                return hugr.std.collections.static_array.StaticArrayVal(
                    vs, elem_ty.to_hugr(ctx), name=f"static_pyarray.{next(tmp_vars)}"
                )
        case None:
            return hugr.val.Unit
        case _:
            return None
    return None


ARRAY_COMPREHENSION_INIT: Final[GlobalConstId] = GlobalConstId.fresh(
    "array.__comprehension.init"
)

ARRAY_UNWRAP_ELEM: Final[GlobalConstId] = GlobalConstId.fresh("array.__unwrap_elem")
ARRAY_WRAP_ELEM: Final[GlobalConstId] = GlobalConstId.fresh("array.__wrap_elem")

ARRAY_READ_BOOL: Final[GlobalConstId] = GlobalConstId.fresh("array.__read_bool")
ARRAY_MAKE_OPAQUE_BOOL: Final[GlobalConstId] = GlobalConstId.fresh(
    "array.__make_opaque_bool"
)


def array_comprehension_init_func(ctx: CompilerContext) -> hf.Function:
    """Returns the Hugr function that is used to initialise arrays elements before a
    comprehension.

    Just returns the `None` variant of the optional element type.

    See https://github.com/CQCL/guppylang/issues/629
    """
    v = ht.Variable(0, ht.TypeBound(ht.TypeBound.Linear))
    sig = ht.PolyFuncType(
        params=[ht.TypeTypeParam(ht.TypeBound.Linear)],
        body=ht.FunctionType([], [ht.Option(v)]),
    )
    func, already_defined = ctx.declare_global_func(ARRAY_COMPREHENSION_INIT, sig)
    if not already_defined:
        func.set_outputs(func.add_op(ops.Tag(0, ht.Option(v))))
    return func


def array_unwrap_elem(ctx: CompilerContext) -> hf.Function:
    """Returns the Hugr function that is used to unwrap the elements in an option array
    to turn it into a regular array."""
    v = ht.Variable(0, ht.TypeBound(ht.TypeBound.Linear))
    sig = ht.PolyFuncType(
        params=[ht.TypeTypeParam(ht.TypeBound.Linear)],
        body=ht.FunctionType([ht.Option(v)], [v]),
    )
    func, already_defined = ctx.declare_global_func(ARRAY_UNWRAP_ELEM, sig)
    if not already_defined:
        msg = "Linear array element has already been used"
        func.set_outputs(build_unwrap(func, func.inputs()[0], msg))
    return func


def array_wrap_elem(ctx: CompilerContext) -> hf.Function:
    """Returns the Hugr function that is used to wrap the elements in an regular array
    to turn it into a option array."""
    v = ht.Variable(0, ht.TypeBound(ht.TypeBound.Linear))
    sig = ht.PolyFuncType(
        params=[ht.TypeTypeParam(ht.TypeBound.Linear)],
        body=ht.FunctionType([v], [ht.Option(v)]),
    )
    func, already_defined = ctx.declare_global_func(ARRAY_WRAP_ELEM, sig)
    if not already_defined:
        func.set_outputs(func.add_op(ops.Tag(1, ht.Option(v)), func.inputs()[0]))
    return func


def array_read_bool(ctx: CompilerContext) -> hf.Function:
    """Returns the Hugr function that is used to unwrap the elements in an option array
    to turn it into a regular array."""
    sig = ht.PolyFuncType(
        params=[],
        body=ht.FunctionType([OpaqueBool], [ht.Bool]),
    )
    func, already_defined = ctx.declare_global_func(ARRAY_READ_BOOL, sig)
    if not already_defined:
        func.set_outputs(func.add_op(read_bool(), func.inputs()[0]))
    return func


def array_make_opaque_bool(ctx: CompilerContext) -> hf.Function:
    """Returns the Hugr function that is used to unwrap the elements in an option array
    to turn it into a regular array."""
    sig = ht.PolyFuncType(
        params=[],
        body=ht.FunctionType([ht.Bool], [OpaqueBool]),
    )
    func, already_defined = ctx.declare_global_func(ARRAY_MAKE_OPAQUE_BOOL, sig)
    if not already_defined:
        func.set_outputs(func.add_op(make_opaque(), func.inputs()[0]))
    return func


T = TypeVar("T")


def doesnt_contain_none(xs: list[T | None]) -> TypeGuard[list[T]]:
    """Checks if a list contains `None`."""
    return all(x is not None for x in xs)


def apply_array_op_with_conversions(
    ctx: CompilerContext,
    builder: DfBase[ops.DfParentOp],
    op: ops.DataflowOp,
    elem_ty: ht.Type,
    size_arg: ht.TypeArg,
    input_array: Wire,
    convert_bool: bool = False,
) -> Wire:
    """Applies common transformations to a Guppy array input before it can be passed to
    a Hugr op operating on a standard Hugr array, and then reverses them again on the
    output array.

    Transformations:
    1. Unwraps / wraps elements in options.
    3. (Optional) Converts from / to opaque bool to / from Hugr bool.
    2. Converts from / to value array to / from standard Hugr array.
    """
    unwrap = array_unwrap_elem(ctx)
    unwrap = builder.load_function(
        unwrap,
        instantiation=ht.FunctionType([ht.Option(elem_ty)], [elem_ty]),
        type_args=[ht.TypeTypeArg(elem_ty)],
    )
    map_op = array_map(ht.Option(elem_ty), size_arg, elem_ty)
    unwrapped_array = builder.add_op(map_op, input_array, unwrap)

    if convert_bool:
        array_read = array_read_bool(ctx)
        array_read = builder.load_function(array_read)
        map_op = array_map(OpaqueBool, size_arg, ht.Bool)
        unwrapped_array = builder.add_op(map_op, unwrapped_array, array_read)
        elem_ty = ht.Bool

    unwrapped_array = builder.add_op(
        array_convert_to_std_array(elem_ty, size_arg), unwrapped_array
    )

    result_array = builder.add_op(op, unwrapped_array)

    result_array = builder.add_op(
        array_convert_from_std_array(elem_ty, size_arg), result_array
    )

    if convert_bool:
        array_make_opaque = array_make_opaque_bool(ctx)
        array_make_opaque = builder.load_function(array_make_opaque)
        map_op = array_map(ht.Bool, size_arg, OpaqueBool)
        result_array = builder.add_op(map_op, result_array, array_make_opaque)
        elem_ty = OpaqueBool

    wrap = array_wrap_elem(ctx)
    wrap = builder.load_function(
        wrap,
        instantiation=ht.FunctionType([elem_ty], [ht.Option(elem_ty)]),
        type_args=[ht.TypeTypeArg(elem_ty)],
    )
    map_op = array_map(elem_ty, size_arg, ht.Option(elem_ty))
    return builder.add_op(map_op, result_array, wrap)



from collections.abc import Callable

from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def foo(x: qubit) -> qubit: ...


@guppy
def test() -> Callable[[qubit @owned], qubit]:
    return foo


test.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit


@guppy.declare
def foo(q: qubit) -> None: ...


@guppy.struct
class MyImmutableContainer:
    q: qubit

    @guppy.declare
    def __getitem__(self: "MyImmutableContainer", idx: int) -> qubit: ...


@guppy
def test(c: MyImmutableContainer) -> MyImmutableContainer:
    foo(c[0])
    return c


test.compile()



from collections.abc import Callable
from typing import Generic

import pytest
from hugr import Hugr, ops

from guppylang import guppy, array
from guppylang.std.num import nat


def funcs_defs(h: Hugr) -> list[str]:
    return [h[node].op.f_name for node in h if isinstance(h[node].op, ops.FuncDefn)]


def test_bool(validate, run_int_fn):
    B = guppy.const_var("B", "bool")

    @guppy.struct
    class Dummy(Generic[B]):
        """Unit struct type to introduce generic B into the signature of `foo` below.

        This is needed because we don't support the `def foo[B: bool]` syntax yet to
        introduce type params that are not referenced in the signature.
        """

    @guppy
    def foo(_: Dummy[B]) -> bool:
        return B

    @guppy
    def main() -> int:
        s = 0
        if foo[True](Dummy()):
            s += 1
        if foo[False](Dummy()):
            s += 10
        if foo[True](Dummy()):
            s += 100
        if foo[False](Dummy()):
            s += 10000
        return s

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 2 monomorphizations of foo (Dummy constructor is inlined)
    assert len(funcs_defs(compiled.modules[0])) == 3

    run_int_fn(main, 101)


@pytest.mark.xfail(reason="https://github.com/CQCL/guppylang/issues/1030")
def test_int(validate):
    IT = guppy.const_var("IT", "int§")

    @guppy.struct
    class Dummy(Generic[IT]):
        """Unit struct type to introduce generic I into the signature of `foo` below.

        This is needed because we don't support the `def foo[I: int]` syntax yet to
        introduce type params that are not referenced in the signature.
        """

    @guppy
    def foo(_: Dummy[IT]) -> float:
        return IT

    @guppy
    def main() -> float:
        return (
            foo[1](Dummy())
            + foo[2](Dummy())
            + foo[-2](Dummy())
            + foo[3](Dummy())
            + foo[1](Dummy())
            + foo[1](Dummy())
        )

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 4 monomorphizations of foo (Dummy constructor is inlined)
    assert len(funcs_defs(compiled.modules[0])) == 5


def test_float(validate, run_float_fn_approx):
    F = guppy.const_var("F", "float")

    @guppy.struct
    class Dummy(Generic[F]):
        """Unit struct type to introduce generic F into the signature of `foo` below.

        This is needed because we don't support the `def foo[F: float]` syntax yet to
        introduce type params that are not referenced in the signature.
        """

    @guppy
    def foo(_: Dummy[F]) -> float:
        return F

    @guppy
    def main() -> float:
        return (
            foo[1.5](Dummy())
            + foo[2.5](Dummy())
            + foo[3.5](Dummy())
            + foo[1.5](Dummy())
            + foo[1.5](Dummy())
        )

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 3 monomorphizations of foo (Dummy constructor is inlined)
    assert len(funcs_defs(compiled.modules[0])) == 4

    run_float_fn_approx(main, 10.5)


@pytest.mark.xfail(reason="https://github.com/CQCL/guppylang/issues/1030")
def test_string(validate):
    S = guppy.const_var("S", "str")

    @guppy.struct
    class Dummy(Generic[S]):
        """Unit struct type to introduce generic S into the signature of `foo` below.

        This is needed because we don't support the `def foo[S: str]` syntax yet to
        introduce type params that are not referenced in the signature.
        """

    @guppy
    def foo(_: Dummy[S]) -> str:
        return S

    @guppy
    def main() -> tuple[str, str, str, str, str]:
        return (
            foo[""](Dummy()),
            foo["a"](Dummy()),
            foo["A"](Dummy()),
            foo["ä"](Dummy()),
            foo["a"](Dummy()),
        )

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 4 monomorphizations of foo (Dummy constructor is inlined)
    assert len(funcs_defs(compiled.modules[0])) == 5


def test_chain(validate, run_int_fn):
    B = guppy.const_var("B", "bool")

    @guppy.struct
    class Dummy(Generic[B]):
        """Unit struct type to introduce generic B into the signatures below.

        This is needed because we don't support the `def foo[B: bool]` syntax yet to
        introduce type params that are not referenced in the signature.
        """

    @guppy
    def a(x: Dummy[B]) -> bool:
        return b(x)

    @guppy
    def b(x: Dummy[B]) -> bool:
        return c(x)

    @guppy
    def c(x: Dummy[B]) -> bool:
        return d(x)

    @guppy
    def d(_: Dummy[B]) -> bool:
        return B

    @guppy
    def main() -> int:
        x = a[True](Dummy())
        b[True](Dummy())
        c[True](Dummy())
        d[True](Dummy())
        return 1 if x else 0

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 4 monomorphizations (a, b, c, d)
    assert len(funcs_defs(compiled.modules[0])) == 5

    run_int_fn(main, 1)


def test_recursion(validate):
    B = guppy.const_var("B", "bool")

    @guppy.struct
    class Dummy(Generic[B]):
        """Unit struct type to introduce generic B into the signatures below.

        This is needed because we don't support the `def foo[B: bool]` syntax yet to
        introduce type params that are not referenced in the signature.
        """

    @guppy
    def foo(_: Dummy[B]) -> int:
        return bar[True](Dummy()) + foo[False](Dummy())

    @guppy
    def bar(_: Dummy[B]) -> int:
        return foo[True](Dummy()) + bar[False](Dummy())

    @guppy
    def baz(d: Dummy[B]) -> int:
        return foo(d)

    @guppy
    def main() -> int:
        return baz[True](Dummy())

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 5 monomorphizations of foo/bar/baz
    assert len(funcs_defs(compiled.modules[0])) == 6


def test_many(validate):
    B = guppy.const_var("B", "bool")
    F = guppy.const_var("F", "float")
    N = guppy.nat_var("N")

    T1 = guppy.type_var("T1")
    T2 = guppy.type_var("T2")
    T3 = guppy.type_var("T3")

    @guppy.struct
    class MyStruct(Generic[T1, B, T2, N, T3, F]):
        """Unit struct type to introduce generics into the signatures below.

        This is needed because we don't support the `def foo[S: str]` syntax yet to
        introduce type params that are not referenced in the signature.
        """

        x1: T1
        x2: T2
        x3s: array[T3, N]

    @guppy.declare
    def bar(xs: array[T1, N]) -> None: ...

    @guppy
    def baz(s: MyStruct[T1, B, T2, N, T3, F]) -> MyStruct[T1, B, T2, N, T3, F]:
        return baz(s)

    @guppy
    def foo(s: MyStruct[int, False, T2, N, T3, F]) -> float:
        bar(s.x3s)
        baz(s)
        return N + F

    @guppy
    def main() -> None:
        s1: MyStruct[int, False, float, 3, bool, 4.2] = MyStruct(
            1, 1.0, array(True, False, True)
        )
        s1 = baz(baz(s1))
        bar(s1.x3s)
        foo(s1)

        s2: MyStruct[int, False, bool, 1, nat, 4.2] = MyStruct(0, False, array(nat(42)))
        s2 = baz(baz(s2))
        bar(s2.x3s)
        foo(s2)

        s3: MyStruct[int, False, bool, 1, float, 1.5] = MyStruct(0, False, array(4.2))
        s3 = baz(baz(s3))
        bar(s3.x3s)
        foo(s3)

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 2 monomorphizations of foo and baz each. Note that `s2`
    # doesn't generate a monomorphisation since it shares the relevant mono-parameters
    # with `s1`
    assert len(funcs_defs(compiled.modules[0])) == 5


def test_constructor(validate):
    B = guppy.const_var("B", "bool")
    F = guppy.const_var("F", "float")

    @guppy.struct
    class MyStruct(Generic[B, F]):
        pass

    @guppy
    def main() -> None:
        s1: MyStruct[True, 1.0] = MyStruct()  # This is inlined
        s2: MyStruct[False, 1.0] = MyStruct()  # This is inlined
        f1 = MyStruct[True, 2.0]  # This is monomorphized
        f2 = MyStruct[False, 2.0]  # This is monomorphized
        f1()
        f2()

    compiled = main.compile()
    validate(compiled)

    # Check we have main, and 2 monomorphizations of the MyStruct constructor
    assert len(funcs_defs(compiled.modules[0])) == 3


def test_higher_order(validate):
    B = guppy.const_var("B", "bool")
    F = guppy.const_var("F", "float")

    @guppy.struct
    class Struct(Generic[B, F]):
        pass

    @guppy
    def fun1(x: Struct[B, F]) -> None:
        pass

    @guppy
    def fun2(x: Struct[True, F]) -> None:
        pass

    @guppy
    def fun3(x: Struct[B, 42.0]) -> None:
        pass

    @guppy
    def foo(f: Callable[[Struct[B, 42.0]], None]) -> None:
        pass

    @guppy
    def main() -> None:
        foo[True](fun1)
        foo[False](fun1)
        foo(fun2)
        foo(fun3[False])
        foo(fun3[True])

    compiled = main.compile()
    validate(compiled)

    # Check we have main, fun2, and 2 monomorphizations of fun1, fun3, and foo each
    assert len(funcs_defs(compiled.modules[0])) == 8



import builtins
import re

import pytest

from guppylang.decorator import guppy
from guppylang.std.builtins import array


def test_float(validate):
    @guppy.comptime
    def test(x: int) -> float:
        # Make sure that the mocked float is indistinguishable from the real deal
        assert isinstance(4.0, float)
        assert isinstance(4.0, builtins.float)
        assert isinstance(float(4), float)
        assert isinstance(float(4), builtins.float)
        assert isinstance(float, type)
        assert issubclass(float, float)
        assert issubclass(float, builtins.float)
        assert issubclass(builtins.float, float)
        assert float == builtins.float  # noqa: E721
        assert not float != builtins.float  # noqa: E721
        assert float.__name__ == builtins.float.__name__
        assert float.__qualname__ == builtins.float.__qualname__

        for v in (True, False, -42, 1.5, "-0.42"):
            assert float(v) == builtins.float(v)

        # But they are not identical
        assert float is not builtins.float

        # In particular, `builtins.float` doesn't work for GuppyObjects, but our mocked
        # version does
        err = re.escape("GuppyObject.__float__ returned non-float (type GuppyObject)")
        with pytest.raises(TypeError, match=err):
            builtins.float(x)
        return float(x)

    validate(test.compile())


def test_int(validate):
    @guppy.comptime
    def test(x: float) -> int:
        # Make sure that the mocked int is indistinguishable from the real deal
        assert isinstance(4, int)
        assert isinstance(4, builtins.int)
        assert isinstance(int(4.0), int)
        assert isinstance(int(4.0), builtins.int)
        assert isinstance(int, type)
        assert issubclass(int, int)
        assert issubclass(int, builtins.int)
        assert issubclass(builtins.int, int)
        assert issubclass(bool, int)
        assert int == builtins.int  # noqa: E721
        assert not int != builtins.int  # noqa: E721
        assert int.__name__ == builtins.int.__name__
        assert int.__qualname__ == builtins.int.__qualname__

        for v in (True, False, 42, "-123"):
            assert int(v) == builtins.int(v)

        # But they are not identical
        assert int is not builtins.int

        # In particular, `builtins.int` doesn't work for GuppyObjects, but our mocked
        # version does
        err = re.escape("__int__ returned non-int (type GuppyObject)")
        with pytest.raises(TypeError, match=err):
            builtins.int(x)
        return int(x)

    validate(test.compile())


def test_len(validate):
    @guppy.struct
    class S:
        x: int

        @guppy
        def __len__(self: "S") -> int:
            return self.x

    @guppy.comptime
    def test(xs: array[int, 10]) -> int:
        # Make sure that the mocked len is indistinguishable from the real deal
        assert len([1, 2, 3, 4, 5]) == 5
        assert len({1, 2}) == 2
        assert len({1: 1}) == 1
        assert len(()) == 0
        assert len(xs) == 10

        # But they are not identical
        assert len != builtins.len
        assert len is not builtins.len

        # In particular, `builtins.len` doesn't work for GuppyObjects, but our mocked
        # version does
        s = S(100)
        err = re.escape("object of type 'GuppyStructObject' has no len()")
        with pytest.raises(TypeError, match=err):
            builtins.len(s)
        return len(s)

    validate(test.compile())



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.declare
def foo(q: qubit, x: int) -> int: ...


@guppy
def test(q: qubit @owned) -> tuple[int, qubit]:
    # This doesn't work since arguments are evaluated from left to right
    return foo(q, foo(q, foo(q, 0))), q


test.compile()



import ast
from abc import ABC, abstractmethod
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, ClassVar

from hugr import Wire, ops
from hugr import tys as ht
from hugr.build.dfg import DfBase

from guppylang_internals.ast_util import (
    AstNode,
    get_type,
    has_empty_body,
    with_loc,
    with_type,
)
from guppylang_internals.checker.core import Context, Globals
from guppylang_internals.checker.expr_checker import check_call, synthesize_call
from guppylang_internals.checker.func_checker import check_signature
from guppylang_internals.compiler.core import (
    CompilerContext,
    DFContainer,
    GlobalConstId,
    partially_monomorphize_args,
)
from guppylang_internals.definition.common import ParsableDef
from guppylang_internals.definition.value import CallReturnWires, CompiledCallableDef
from guppylang_internals.diagnostic import Error, Help
from guppylang_internals.error import GuppyError, InternalGuppyError
from guppylang_internals.nodes import GlobalCall
from guppylang_internals.span import SourceMap
from guppylang_internals.std._internal.compiler.tket_bool import (
    OpaqueBool,
    make_opaque,
    read_bool,
)
from guppylang_internals.tys.subst import Inst, Subst
from guppylang_internals.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    Type,
    type_to_row,
)

if TYPE_CHECKING:
    from guppylang_internals.definition.function import PyFunc


@dataclass(frozen=True)
class BodyNotEmptyError(Error):
    title: ClassVar[str] = "Unexpected function body"
    span_label: ClassVar[str] = "Body of custom function `{name}` must be empty"
    name: str


@dataclass(frozen=True)
class NoSignatureError(Error):
    title: ClassVar[str] = "Type signature missing"
    span_label: ClassVar[str] = "Custom function `{name}` requires a type signature"
    name: str

    @dataclass(frozen=True)
    class Suggestion(Help):
        message: ClassVar[str] = (
            "Annotate the type signature of `{name}` or disallow the use of `{name}` "
            "as a higher-order value: `@custom_function(..., higher_order_value=False)`"
        )

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(NoSignatureError.Suggestion(None))


@dataclass(frozen=True)
class NotHigherOrderError(Error):
    title: ClassVar[str] = "Not higher-order"
    span_label: ClassVar[str] = (
        "Function `{name}` may not be used as a higher-order value"
    )
    name: str


@dataclass(frozen=True)
class RawCustomFunctionDef(ParsableDef):
    """A raw custom function definition provided by the user.

    Custom functions provide their own checking and compilation logic using a
    `CustomCallChecker` and a `CustomCallCompiler`.

    The raw definition stores exactly what the user has written (i.e. the AST together
    with the provided checker and compiler), without inspecting the signature.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
        call_checker: The custom call checker.
        call_compiler: The custom call compiler.
        higher_order_value: Whether the function may be used as a higher-order value.
        signature: User-provided signature.
    """

    python_func: "PyFunc"
    call_checker: "CustomCallChecker"
    call_compiler: "CustomInoutCallCompiler"

    # Whether the function may be used as a higher-order value. This is only possible
    # if a static type for the function is provided.
    higher_order_value: bool

    signature: FunctionType | None

    description: str = field(default="function", init=False)

    def parse(self, globals: "Globals", sources: SourceMap) -> "CustomFunctionDef":
        """Parses and checks the signature of the custom function.

        The signature is optional if custom type checking logic is provided by the user.
        However, note that a signature must be provided by either annotation or as an
        argument, if we want to use the function as a higher-order value. If a signature
        is provided as an argument, this will override any annotation.

        If no signature is provided, we fill in the dummy signature `() -> ()`. This
        type will never be inspected, since we rely on the provided custom checking
        code. The only information we need to access is that it's a function type and
        that there are no unsolved existential vars.
        """
        from guppylang_internals.definition.function import parse_py_func

        func_ast, docstring = parse_py_func(self.python_func, sources)
        if not has_empty_body(func_ast):
            raise GuppyError(BodyNotEmptyError(func_ast.body[0], self.name))
        sig = self.signature or self._get_signature(func_ast, globals)
        ty = sig or FunctionType([], NoneType())
        return CustomFunctionDef(
            self.id,
            self.name,
            func_ast,
            ty,
            self.call_checker,
            self.call_compiler,
            self.higher_order_value,
            GlobalConstId.fresh(self.name),
            sig is not None,
        )

    def _get_signature(
        self, node: ast.FunctionDef, globals: Globals
    ) -> FunctionType | None:
        """Returns the type of the function, if known.

        Type annotations are needed if we rely on the default call checker or
        want to allow the usage of the function as a higher-order value.

        Some function types like python's `int()` cannot be expressed in the Guppy
        type system, so we return `None` here and rely on the specialized compiler
        to handle the call.
        """
        requires_type_annotation = (
            isinstance(self.call_checker, DefaultCallChecker) or self.higher_order_value
        )
        has_type_annotation = node.returns or any(
            arg.annotation for arg in node.args.args
        )

        if requires_type_annotation and not has_type_annotation:
            raise GuppyError(NoSignatureError(node, self.name))

        if requires_type_annotation:
            return check_signature(node, globals)
        else:
            return None


@dataclass(frozen=True)
class CustomFunctionDef(CompiledCallableDef):
    """A custom function with parsed and checked signature.

    Args:
        id: The unique definition identifier.
        name: The name of the definition.
        defined_at: The AST node where the definition was defined.
        ty: The type of the function. This may be a dummy value if `has_signature` is
            false.
        call_checker: The custom call checker.
        call_compiler: The custom call compiler.
        higher_order_value: Whether the function may be used as a higher-order value.
        has_signature: Whether the function has a declared signature.
    """

    defined_at: AstNode | None
    ty: FunctionType
    call_checker: "CustomCallChecker"
    call_compiler: "CustomInoutCallCompiler"
    higher_order_value: bool
    higher_order_func_id: GlobalConstId
    has_signature: bool

    description: str = field(default="function", init=False)

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type.

        This is done by invoking the provided `CustomCallChecker`.
        """
        self.call_checker._setup(ctx, node, self)
        new_node, subst = self.call_checker.check(args, ty)
        return with_type(ty, with_loc(node, new_node)), subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: "Context"
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call.

        This is done by invoking the provided `CustomCallChecker`.
        """
        self.call_checker._setup(ctx, node, self)
        new_node, ty = self.call_checker.synthesize(args)
        return with_type(ty, with_loc(node, new_node)), ty

    def load_with_args(
        self,
        type_args: Inst,
        dfg: "DFContainer",
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the custom function as a value into a local dataflow graph.

        This will place a `FunctionDef` node in the local DFG, and load with a
        `LoadFunc` node. This operation will fail if the function is not allowed
        to be used as a higher-order value.
        """
        # TODO: This should be raised during checking, not compilation!
        if not self.higher_order_value:
            raise GuppyError(NotHigherOrderError(node, self.name))
        assert len(self.ty.params) == len(type_args)

        # Partially monomorphize the function if required
        mono_args, rem_args = partially_monomorphize_args(
            self.ty.params, type_args, ctx
        )

        # We create a generic `FunctionDef` that takes some inputs, compiles a call to
        # the function, and returns the results
        func, already_defined = ctx.declare_global_func(
            self.higher_order_func_id,
            self.ty.instantiate_partial(mono_args).to_hugr_poly(ctx),
            mono_args,
        )
        if not already_defined:
            with ctx.set_monomorphized_args(mono_args):
                func_dfg = DFContainer(func, ctx, dfg.locals.copy())
                args: list[Wire] = list(func.inputs())
                generic_ty_args = [param.to_bound() for param in self.ty.params]
                returns = self.compile_call(args, generic_ty_args, func_dfg, ctx, node)
                func.set_outputs(*returns.regular_returns, *returns.inout_returns)

        # Finally, load the function into the local DFG
        mono_ty = self.ty.instantiate(type_args).to_hugr(ctx)
        hugr_ty_args = [ta.to_hugr(ctx) for ta in rem_args]
        return dfg.builder.load_function(func, mono_ty, hugr_ty_args)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: "DFContainer",
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        if self.has_signature:
            concrete_ty = self.ty.instantiate(type_args)
        else:
            assert isinstance(node, GlobalCall)
            concrete_ty = FunctionType(
                [FuncInput(get_type(arg), InputFlags.NoFlags) for arg in node.args],
                get_type(node),
            )
        hugr_ty = concrete_ty.to_hugr(ctx)

        self.call_compiler._setup(type_args, dfg, ctx, node, hugr_ty, self)
        return self.call_compiler.compile_with_inouts(args)


class CustomCallChecker(ABC):
    """Abstract base class for custom function call type checkers."""

    ctx: Context
    node: AstNode
    func: CustomFunctionDef

    def _setup(self, ctx: Context, node: AstNode, func: CustomFunctionDef) -> None:
        self.ctx = ctx
        self.node = node
        self.func = func

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        """Checks the return value against a given type.

        Returns a (possibly) transformed and annotated AST node for the call.
        """
        from guppylang_internals.checker.expr_checker import check_type_against

        expr, res_ty = self.synthesize(args)
        expr, subst, _ = check_type_against(res_ty, ty, expr, self.ctx)
        return expr, subst

    @abstractmethod
    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        """Synthesizes a type for the return value of a call.

        Also returns a (possibly) transformed and annotated argument list.
        """


class CustomInoutCallCompiler(ABC):
    """Abstract base class for custom function call compilers with borrowed args.

    Args:
        builder: The function builder where the function should be defined.
        type_args: The type arguments for the function.
        globals: The compiled globals.
        node: The AST node where the function is defined.
        ty: The type of the function, if known.
    """

    dfg: DFContainer
    type_args: Inst
    ctx: CompilerContext
    node: AstNode
    ty: ht.FunctionType
    func: CustomFunctionDef | None

    def _setup(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
        hugr_ty: ht.FunctionType,
        func: CustomFunctionDef | None,
    ) -> None:
        self.type_args = type_args
        self.dfg = dfg
        self.ctx = ctx
        self.node = node
        self.ty = hugr_ty
        self.func = func

    @abstractmethod
    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        """Compiles a custom function call.

        Returns the outputs of the call together with any borrowed arguments that are
        passed through the function.
        """

    @property
    def builder(self) -> DfBase[ops.DfParentOp]:
        """The hugr dataflow builder."""
        return self.dfg.builder


class CustomCallCompiler(CustomInoutCallCompiler, ABC):
    """Abstract base class for custom function call compilers with only owned args."""

    @abstractmethod
    def compile(self, args: list[Wire]) -> list[Wire]:
        """Compiles a custom function call and returns the resulting ports.

        Use the provided `self.builder` to add nodes to the Hugr graph.
        """

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        return CallReturnWires(self.compile(args), inout_returns=[])


class DefaultCallChecker(CustomCallChecker):
    """Checks function calls by comparing to a type signature."""

    def check(self, args: list[ast.expr], ty: Type) -> tuple[ast.expr, Subst]:
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.func.ty, args, ty, self.node, self.ctx)
        return GlobalCall(def_id=self.func.id, args=args, type_args=inst), subst

    def synthesize(self, args: list[ast.expr]) -> tuple[ast.expr, Type]:
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.func.ty, args, self.node, self.ctx)
        return GlobalCall(def_id=self.func.id, args=args, type_args=inst), ty


class NotImplementedCallCompiler(CustomCallCompiler):
    """Call compiler for custom functions that are already lowered during checking.

    For example, the custom checker could replace the call with a series of calls to
    other functions. In that case, the original function will no longer be present and
    thus doesn't need to be compiled.
    """

    def compile(self, args: list[Wire]) -> list[Wire]:
        raise InternalGuppyError("Function should have been removed during checking")


class OpCompiler(CustomInoutCallCompiler):
    """Call compiler for functions that are directly implemented via Hugr ops.

    args:
        op: A function that takes an instantiation of the type arguments as well as
            the monomorphic function type, and returns a concrete HUGR op.
    """

    op: Callable[[ht.FunctionType, Inst, CompilerContext], ops.DataflowOp]

    def __init__(
        self, op: Callable[[ht.FunctionType, Inst, CompilerContext], ops.DataflowOp]
    ) -> None:
        self.op = op

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        op = self.op(self.ty, self.type_args, self.ctx)
        node = self.builder.add_op(op, *args)
        num_returns = (
            len(type_to_row(self.func.ty.output)) if self.func else len(self.ty.output)
        )
        return CallReturnWires(
            regular_returns=list(node[:num_returns]),
            inout_returns=list(node[num_returns:]),
        )


class BoolOpCompiler(CustomInoutCallCompiler):
    """Call compiler for functions that are directly implemented via Hugr ops but need
    input and/or output conversions from hugr sum bools to the opaque bools Guppy is
    using.

    args:
        op: A function that takes an instantiation of the type arguments as well as
            the monomorphic function type, and returns a concrete HUGR op.
    """

    op: Callable[[ht.FunctionType, Inst, CompilerContext], ops.DataflowOp]

    def __init__(
        self, op: Callable[[ht.FunctionType, Inst, CompilerContext], ops.DataflowOp]
    ) -> None:
        self.op = op

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        converted_in = [ht.Bool if inp == OpaqueBool else inp for inp in self.ty.input]
        converted_out = [
            ht.Bool if out == OpaqueBool else out for out in self.ty.output
        ]
        hugr_op_ty = ht.FunctionType(converted_in, converted_out)
        op = self.op(hugr_op_ty, self.type_args, self.ctx)
        converted_args = [
            self.builder.add_op(read_bool(), arg)
            if self.builder.hugr.port_type(arg.out_port()) == OpaqueBool
            else arg
            for arg in args
        ]
        node = self.builder.add_op(op, *converted_args)
        result = list(node.outputs())
        converted_result = [
            self.builder.add_op(make_opaque(), res)
            if self.builder.hugr.port_type(res.out_port()) == ht.Bool
            else res
            for res in result
        ]
        return CallReturnWires(
            regular_returns=converted_result,
            inout_returns=[],
        )


class NoopCompiler(CustomCallCompiler):
    """Call compiler for functions that are noops."""

    def compile(self, args: list[Wire]) -> list[Wire]:
        return args


class CopyInoutCompiler(CustomInoutCallCompiler):
    """Call compiler for functions that are noops but only want to borrow arguments."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        return CallReturnWires(regular_returns=args, inout_returns=args)



from guppylang.decorator import custom_guppy_decorator, guppy


@custom_guppy_decorator
def my_guppy(f):
    return guppy(f)


def test_custom_decorator(validate):
    def make():
        @my_guppy
        def foooo() -> int:
            return 1 + bar()

        return foooo

    foo = make()

    @guppy
    def bar() -> int:
        return 2 + foo()

    @my_guppy
    def main() -> int:
        return foo() + bar() + main()

    validate(main.compile())


def test_nested(validate):
    def my_guppy():
        @custom_guppy_decorator
        def inner(f):
            return guppy(f)

        return inner

    def make():
        @my_guppy()
        def foooo() -> int:
            return 1 + bar()

        return foooo

    foo = make()

    @guppy
    def bar() -> int:
        return 2 + foo()

    @my_guppy()
    def main() -> int:
        return foo() + bar() + main()

    validate(main.compile())


def test_method(validate):
    class MyGuppy:
        @custom_guppy_decorator
        def decorator(self, f):
            return guppy(f)

    my_guppy = MyGuppy()

    def make():
        @my_guppy.decorator
        def foooo() -> int:
            return 1 + bar()

        return foooo

    foo = make()

    @guppy
    def bar() -> int:
        return 2 + foo()

    @my_guppy.decorator
    def main() -> int:
        return foo() + bar() + main()

    validate(main.compile())



import ast
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Any

import hugr.build.function as hf
import hugr.tys as ht
from hugr import Node, Wire
from hugr.build.dfg import DefinitionBuilder, OpVar

from guppylang_internals.ast_util import AstNode, with_loc
from guppylang_internals.checker.core import Context, Globals
from guppylang_internals.checker.errors.generic import UnsupportedError
from guppylang_internals.checker.expr_checker import (
    check_call,
    synthesize_call,
)
from guppylang_internals.checker.func_checker import (
    check_signature,
)
from guppylang_internals.compiler.core import CompilerContext, DFContainer
from guppylang_internals.definition.common import (
    CompilableDef,
    ParsableDef,
)
from guppylang_internals.definition.function import parse_py_func
from guppylang_internals.definition.value import (
    CallableDef,
    CallReturnWires,
    CompiledCallableDef,
    CompiledHugrNodeDef,
)
from guppylang_internals.error import GuppyError
from guppylang_internals.nodes import GlobalCall
from guppylang_internals.span import SourceMap
from guppylang_internals.tys.subst import Inst, Subst
from guppylang_internals.tys.ty import FunctionType, Type, type_to_row

PyFunc = Callable[..., Any]


@dataclass(frozen=True)
class RawTracedFunctionDef(ParsableDef):
    python_func: PyFunc

    description: str = field(default="function", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "TracedFunctionDef":
        """Parses and checks the user-provided signature of the function."""
        func_ast, _docstring = parse_py_func(self.python_func, sources)
        ty = check_signature(func_ast, globals)
        if ty.parametrized:
            raise GuppyError(UnsupportedError(func_ast, "Generic comptime functions"))
        return TracedFunctionDef(self.id, self.name, func_ast, ty, self.python_func)


@dataclass(frozen=True)
class TracedFunctionDef(RawTracedFunctionDef, CallableDef, CompilableDef):
    python_func: PyFunc
    ty: FunctionType
    defined_at: ast.FunctionDef

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.ty, args, ty, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call."""
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.ty, args, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, ty

    def compile_outer(
        self, module: DefinitionBuilder[OpVar], ctx: CompilerContext
    ) -> "CompiledTracedFunctionDef":
        """Adds a Hugr `FuncDefn` node for this function to the Hugr.

        Note that we don't compile the function body at this point since we don't have
        access to the other compiled functions yet. The body is compiled later in
        `CompiledFunctionDef.compile_inner()`.
        """
        func_type = self.ty.to_hugr_poly(ctx)
        func_def = module.module_root_builder().define_function(
            self.name, func_type.body.input, func_type.body.output, func_type.params
        )
        return CompiledTracedFunctionDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.python_func,
            func_def,
        )


@dataclass(frozen=True)
class CompiledTracedFunctionDef(
    TracedFunctionDef, CompiledCallableDef, CompiledHugrNodeDef
):
    func_def: hf.Function

    @property
    def hugr_node(self) -> Node:
        """The Hugr node this definition was compiled into."""
        return self.func_def.parent_node

    def load_with_args(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the function as a value into a local Hugr dataflow graph."""
        func_ty: ht.FunctionType = self.ty.instantiate(type_args).to_hugr(ctx)
        type_args: list[ht.TypeArg] = [arg.to_hugr(ctx) for arg in type_args]
        return dfg.builder.load_function(self.func_def, func_ty, type_args)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        func_ty: ht.FunctionType = self.ty.instantiate(type_args).to_hugr(ctx)
        type_args: list[ht.TypeArg] = [arg.to_hugr(ctx) for arg in type_args]
        num_returns = len(type_to_row(self.ty.output))
        call = dfg.builder.call(
            self.func_def, *args, instantiation=func_ty, type_args=type_args
        )
        return CallReturnWires(
            regular_returns=list(call[:num_returns]),
            inout_returns=list(call[num_returns:]),
        )

    def compile_inner(self, ctx: CompilerContext) -> None:
        """Compiles the body of the function by tracing it."""
        from guppylang_internals.tracing.function import trace_function

        trace_function(self.python_func, self.ty, self.func_def, ctx, self.defined_at)



from typing import TYPE_CHECKING

from guppylang_internals.ast_util import AstNode
from guppylang_internals.checker.errors.wasm import (
    FirstArgNotModule,
    UnWasmableType,
)
from guppylang_internals.definition.custom import (
    CustomFunctionDef,
    RawCustomFunctionDef,
)
from guppylang_internals.error import GuppyError
from guppylang_internals.span import SourceMap
from guppylang_internals.tys.builtin import wasm_module_info
from guppylang_internals.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    TupleType,
    Type,
)

if TYPE_CHECKING:
    from guppylang_internals.checker.core import Globals


class RawWasmFunctionDef(RawCustomFunctionDef):
    def sanitise_type(self, loc: AstNode | None, fun_ty: FunctionType) -> None:
        # Place to highlight in error messages
        match fun_ty.inputs[0]:
            case FuncInput(ty=ty, flags=InputFlags.Inout) if wasm_module_info(
                ty
            ) is not None:
                pass
            case FuncInput(ty=ty):
                raise GuppyError(FirstArgNotModule(loc, ty))
        for inp in fun_ty.inputs[1:]:
            if not self.is_type_wasmable(inp.ty):
                raise GuppyError(UnWasmableType(loc, inp.ty))
        if not self.is_type_wasmable(fun_ty.output):
            match fun_ty.output:
                case NoneType():
                    pass
                case _:
                    raise GuppyError(UnWasmableType(loc, fun_ty.output))

    def is_type_wasmable(self, ty: Type) -> bool:
        match ty:
            case NumericType():
                return True
            case TupleType(element_types=tys):
                return all(self.is_type_wasmable(ty) for ty in tys)

        return False

    def parse(self, globals: "Globals", sources: SourceMap) -> "CustomFunctionDef":
        parsed = super().parse(globals, sources)
        self.sanitise_type(parsed.defined_at, parsed.ty)
        return parsed



import ast
import inspect
import linecache
import sys
from collections.abc import Sequence
from dataclasses import dataclass
from types import FrameType
from typing import ClassVar

from hugr import Wire, ops

from guppylang_internals.ast_util import AstNode, annotate_location
from guppylang_internals.checker.core import Globals
from guppylang_internals.checker.errors.generic import (
    ExpectedError,
    UnexpectedError,
    UnsupportedError,
)
from guppylang_internals.compiler.core import GlobalConstId
from guppylang_internals.definition.common import (
    CheckableDef,
    CompiledDef,
    DefId,
    ParsableDef,
    UnknownSourceError,
)
from guppylang_internals.definition.custom import (
    CustomCallCompiler,
    CustomFunctionDef,
    DefaultCallChecker,
)
from guppylang_internals.definition.function import parse_source
from guppylang_internals.definition.parameter import ParamDef
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.diagnostic import Error, Help, Note
from guppylang_internals.engine import DEF_STORE
from guppylang_internals.error import GuppyError, InternalGuppyError
from guppylang_internals.ipython_inspect import is_running_ipython
from guppylang_internals.span import SourceMap, Span, to_span
from guppylang_internals.tys.arg import Argument
from guppylang_internals.tys.param import Parameter, check_all_args
from guppylang_internals.tys.parsing import type_from_ast
from guppylang_internals.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    StructType,
    Type,
)

if sys.version_info >= (3, 12):
    from guppylang_internals.tys.parsing import parse_parameter


@dataclass(frozen=True)
class UncheckedStructField:
    """A single field on a struct whose type has not been checked yet."""

    name: str
    type_ast: ast.expr


@dataclass(frozen=True)
class StructField:
    """A single field on a struct."""

    name: str
    ty: Type


@dataclass(frozen=True)
class RedundantParamsError(Error):
    title: ClassVar[str] = "Generic parameters already specified"
    span_label: ClassVar[str] = "Duplicate specification of generic parameters"
    struct_name: str

    @dataclass(frozen=True)
    class PrevSpec(Note):
        span_label: ClassVar[str] = (
            "Parameters of `{struct_name}` are already specified here"
        )


@dataclass(frozen=True)
class DuplicateFieldError(Error):
    title: ClassVar[str] = "Duplicate field"
    span_label: ClassVar[str] = (
        "Struct `{struct_name}` already contains a field named `{field_name}`"
    )
    struct_name: str
    field_name: str


@dataclass(frozen=True)
class NonGuppyMethodError(Error):
    title: ClassVar[str] = "Not a Guppy method"
    span_label: ClassVar[str] = (
        "Method `{method_name}` of struct `{struct_name}` is not a Guppy function"
    )
    struct_name: str
    method_name: str

    @dataclass(frozen=True)
    class Suggestion(Help):
        message: ClassVar[str] = (
            "Add a `@guppy` annotation to turn `{method_name}` into a Guppy method"
        )

    def __post_init__(self) -> None:
        self.add_sub_diagnostic(NonGuppyMethodError.Suggestion(None))


@dataclass(frozen=True)
class RawStructDef(TypeDef, ParsableDef):
    """A raw struct type definition that has not been parsed yet."""

    python_class: type

    def parse(self, globals: Globals, sources: SourceMap) -> "ParsedStructDef":
        """Parses the raw class object into an AST and checks that it is well-formed."""
        frame = DEF_STORE.frames[self.id]
        cls_def = parse_py_class(self.python_class, frame, sources)
        if cls_def.keywords:
            raise GuppyError(UnexpectedError(cls_def.keywords[0], "keyword"))

        # Look for generic parameters from Python 3.12 style syntax
        params = []
        params_span: Span | None = None
        if sys.version_info >= (3, 12):
            if cls_def.type_params:
                first, last = cls_def.type_params[0], cls_def.type_params[-1]
                params_span = Span(to_span(first).start, to_span(last).end)
                params = [
                    parse_parameter(node, idx, globals)
                    for idx, node in enumerate(cls_def.type_params)
                ]

        # The only base we allow is `Generic[...]` to specify generic parameters with
        # the legacy syntax
        match cls_def.bases:
            case []:
                pass
            case [base] if elems := try_parse_generic_base(base):
                # Complain if we already have Python 3.12 generic params
                if params_span is not None:
                    err: Error = RedundantParamsError(base, self.name)
                    err.add_sub_diagnostic(RedundantParamsError.PrevSpec(params_span))
                    raise GuppyError(err)
                params = params_from_ast(elems, globals)
            case bases:
                err = UnsupportedError(bases[0], "Struct inheritance", singular=True)
                raise GuppyError(err)

        fields: list[UncheckedStructField] = []
        used_field_names: set[str] = set()
        used_func_names: dict[str, ast.FunctionDef] = {}
        for i, node in enumerate(cls_def.body):
            match i, node:
                # We allow `pass` statements to define empty structs
                case _, ast.Pass():
                    pass
                # Docstrings are also fine if they occur at the start
                case 0, ast.Expr(value=ast.Constant(value=v)) if isinstance(v, str):
                    pass
                # Ensure that all function definitions are Guppy functions
                case _, ast.FunctionDef(name=name) as node:
                    from guppylang.defs import GuppyDefinition

                    v = getattr(self.python_class, name)
                    if not isinstance(v, GuppyDefinition):
                        raise GuppyError(NonGuppyMethodError(node, self.name, name))
                    used_func_names[name] = node
                    if name in used_field_names:
                        raise GuppyError(DuplicateFieldError(node, self.name, name))
                # Struct fields are declared via annotated assignments without value
                case _, ast.AnnAssign(target=ast.Name(id=field_name)) as node:
                    if node.value:
                        err = UnsupportedError(node.value, "Default struct values")
                        raise GuppyError(err)
                    if field_name in used_field_names:
                        err = DuplicateFieldError(node.target, self.name, field_name)
                        raise GuppyError(err)
                    fields.append(UncheckedStructField(field_name, node.annotation))
                    used_field_names.add(field_name)
                case _, node:
                    err = UnexpectedError(
                        node, "statement", unexpected_in="struct definition"
                    )
                    raise GuppyError(err)

        # Ensure that functions don't override struct fields
        if overridden := used_field_names.intersection(used_func_names.keys()):
            x = overridden.pop()
            raise GuppyError(DuplicateFieldError(used_func_names[x], self.name, x))

        return ParsedStructDef(self.id, self.name, cls_def, params, fields)

    def check_instantiate(
        self, args: Sequence[Argument], loc: AstNode | None = None
    ) -> Type:
        raise InternalGuppyError("Tried to instantiate raw struct definition")


@dataclass(frozen=True)
class ParsedStructDef(TypeDef, CheckableDef):
    """A struct definition whose fields have not been checked yet."""

    defined_at: ast.ClassDef
    params: Sequence[Parameter]
    fields: Sequence[UncheckedStructField]

    def check(self, globals: Globals) -> "CheckedStructDef":
        """Checks that all struct fields have valid types."""
        # Before checking the fields, make sure that this definition is not recursive,
        # otherwise the code below would not terminate.
        # TODO: This is not ideal (see todo in `check_instantiate`)
        param_var_mapping = {p.name: p for p in self.params}
        check_not_recursive(self, globals, param_var_mapping)

        fields = [
            StructField(f.name, type_from_ast(f.type_ast, globals, param_var_mapping))
            for f in self.fields
        ]
        return CheckedStructDef(
            self.id, self.name, self.defined_at, self.params, fields
        )

    def check_instantiate(
        self, args: Sequence[Argument], loc: AstNode | None = None
    ) -> Type:
        """Checks if the struct can be instantiated with the given arguments."""
        check_all_args(self.params, args, self.name, loc)
        # Obtain a checked version of this struct definition so we can construct a
        # `StructType` instance
        globals = Globals(DEF_STORE.frames[self.id])
        # TODO: This is quite bad: If we have a cyclic definition this will not
        #  terminate, so we have to check for cycles in every call to `check`. The
        #  proper way to deal with this is changing `StructType` such that it only
        #  takes a `DefId` instead of a `CheckedStructDef`. But this will be a bigger
        #  refactor...
        checked_def = self.check(globals)
        return StructType(args, checked_def)


@dataclass(frozen=True)
class CheckedStructDef(TypeDef, CompiledDef):
    """A struct definition that has been fully checked."""

    defined_at: ast.ClassDef
    params: Sequence[Parameter]
    fields: Sequence[StructField]

    def check_instantiate(
        self, args: Sequence[Argument], loc: AstNode | None = None
    ) -> Type:
        """Checks if the struct can be instantiated with the given arguments."""
        check_all_args(self.params, args, self.name, loc)
        return StructType(args, self)

    def generated_methods(self) -> list[CustomFunctionDef]:
        """Auto-generated methods for this struct."""

        class ConstructorCompiler(CustomCallCompiler):
            """Compiler for the `__new__` constructor method of a struct."""

            def compile(self, args: list[Wire]) -> list[Wire]:
                return list(self.builder.add(ops.MakeTuple()(*args)))

        constructor_sig = FunctionType(
            inputs=[
                FuncInput(f.ty, InputFlags.Owned if f.ty.linear else InputFlags.NoFlags)
                for f in self.fields
            ],
            output=StructType(
                defn=self, args=[p.to_bound(i) for i, p in enumerate(self.params)]
            ),
            input_names=[f.name for f in self.fields],
            params=self.params,
        )
        constructor_def = CustomFunctionDef(
            id=DefId.fresh(),
            name="__new__",
            defined_at=self.defined_at,
            ty=constructor_sig,
            call_checker=DefaultCallChecker(),
            call_compiler=ConstructorCompiler(),
            higher_order_value=True,
            higher_order_func_id=GlobalConstId.fresh(f"{self.name}.__new__"),
            has_signature=True,
        )
        return [constructor_def]


def parse_py_class(
    cls: type, defining_frame: FrameType, sources: SourceMap
) -> ast.ClassDef:
    """Parses a Python class object into an AST."""
    module = inspect.getmodule(cls)
    if module is None:
        raise GuppyError(UnknownSourceError(None, cls))

    # If we are running IPython, `inspect.getsourcefile` won't work if the class was
    # defined inside a cell. See
    #  - https://bugs.python.org/issue33826
    #  - https://github.com/ipython/ipython/issues/11249
    #  - https://github.com/wandb/weave/pull/1864
    if is_running_ipython() and module.__name__ == "__main__":
        file: str | None = defining_frame.f_code.co_filename
    else:
        file = inspect.getsourcefile(cls)
    if file is None:
        raise GuppyError(UnknownSourceError(None, cls))

    # We can't rely on `inspect.getsourcelines` since it doesn't work properly for
    # classes prior to Python 3.13. See https://github.com/CQCL/guppylang/issues/1107.
    # Instead, we reproduce the behaviour of Python >= 3.13 using the `__firstlineno__`
    # attribute. See https://github.com/python/cpython/blob/3.13/Lib/inspect.py#L1052.
    # In the decorator, we make sure that `__firstlineno__` is set, even if we're not
    # on Python 3.13.
    file_lines = linecache.getlines(file)
    line_offset = cls.__firstlineno__  # type: ignore[attr-defined]
    source_lines = inspect.getblock(file_lines[line_offset - 1 :])
    source, cls_ast, line_offset = parse_source(source_lines, line_offset)

    # Store the source file in our cache
    sources.add_file(file)
    annotate_location(cls_ast, source, file, line_offset)
    if not isinstance(cls_ast, ast.ClassDef):
        raise GuppyError(ExpectedError(cls_ast, "a class definition"))
    return cls_ast


def try_parse_generic_base(node: ast.expr) -> list[ast.expr] | None:
    """Checks if an AST node corresponds to a `Generic[T1, ..., Tn]` base class.

    Returns the generic parameters or `None` if the AST has a different shape
    """
    match node:
        case ast.Subscript(value=ast.Name(id="Generic"), slice=elem):
            return elem.elts if isinstance(elem, ast.Tuple) else [elem]
        case _:
            return None


@dataclass(frozen=True)
class RepeatedTypeParamError(Error):
    title: ClassVar[str] = "Duplicate type parameter"
    span_label: ClassVar[str] = "Type parameter `{name}` cannot be used multiple times"
    name: str


def params_from_ast(nodes: Sequence[ast.expr], globals: Globals) -> list[Parameter]:
    """Parses a list of AST nodes into unique type parameters.

    Raises user errors if the AST nodes don't correspond to parameters or parameters
    occur multiple times.
    """
    params: list[Parameter] = []
    params_set: set[DefId] = set()
    for node in nodes:
        if isinstance(node, ast.Name) and node.id in globals:
            defn = globals[node.id]
            if isinstance(defn, ParamDef):
                if defn.id in params_set:
                    raise GuppyError(RepeatedTypeParamError(node, node.id))
                params.append(defn.to_param(len(params)))
                params_set.add(defn.id)
                continue
        raise GuppyError(ExpectedError(node, "a type parameter"))
    return params


def check_not_recursive(
    defn: ParsedStructDef, globals: Globals, param_var_mapping: dict[str, Parameter]
) -> None:
    """Throws a user error if the given struct definition is recursive."""

    # TODO: The implementation below hijacks the type parsing logic to detect recursive
    #  structs. This is not great since it repeats the work done during checking. We can
    #  get rid of this after resolving the todo in `ParsedStructDef.check_instantiate()`

    def dummy_check_instantiate(
        args: Sequence[Argument],
        loc: AstNode | None = None,
    ) -> Type:
        raise GuppyError(UnsupportedError(loc, "Recursive structs"))

    original = defn.check_instantiate
    object.__setattr__(defn, "check_instantiate", dummy_check_instantiate)
    for fld in defn.fields:
        type_from_ast(fld.type_ast, globals, param_var_mapping)
    object.__setattr__(defn, "check_instantiate", original)



from guppylang.decorator import guppy
from guppylang.std.builtins import array


@guppy.comptime
def test(xs: array[int, 10]) -> None:
    xs.clear()
    for i in range(10):
        xs.append(float(i))


test.compile()



from guppylang.decorator import guppy


def test_docstring(validate):
    @guppy
    def f() -> None:
        "Example docstring"

    @guppy
    def g() -> None:
        """Multi
        line
        doc
        string.
        """
        f()

    @guppy
    def nested() -> None:
        """Test docstrings in nested functions"""

        def f_nested() -> None:
            "Example docstring"
            f_nested()
            g()

        def g_nested() -> None:
            """Multi
            line
            doc
            string.
            """

    validate(f.compile())
    validate(g.compile())
    validate(nested.compile())



from guppylang.decorator import guppy


@guppy.struct
class S:
    x: int


@guppy.comptime
def foo(s: S) -> None:
    s.x = 1


@guppy.comptime
def main() -> None:
    s = S(0)
    foo(s)
    # Python users would expect s to be mutated by the call above. However, we cannot
    # provide this semantics, so the mutation in `foo` is rejected by the compiler.


main.compile()



from guppylang import guppy
from tests.util import compile_guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned

@guppy.declare
def use(x: qubit @ owned) -> None: ...

@guppy.struct
class S:
    t: tuple[qubit, int]

@compile_guppy
def foo(b: bool) -> S:
    s = S((qubit(), 1))
    use(s.t[0])
    return s


import ast
import inspect
from collections.abc import Callable, Sequence
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any

import hugr.build.function as hf
import hugr.tys as ht
from hugr import Node, Wire
from hugr.build.dfg import DefinitionBuilder, OpVar
from hugr.hugr.node_port import ToNode

from guppylang_internals.ast_util import AstNode, annotate_location, with_loc, with_type
from guppylang_internals.checker.cfg_checker import CheckedCFG
from guppylang_internals.checker.core import Context, Globals, Place
from guppylang_internals.checker.errors.generic import ExpectedError
from guppylang_internals.checker.expr_checker import check_call, synthesize_call
from guppylang_internals.checker.func_checker import (
    check_global_func_def,
    check_signature,
    parse_function_with_docstring,
)
from guppylang_internals.compiler.core import (
    CompilerContext,
    DFContainer,
    PartiallyMonomorphizedArgs,
)
from guppylang_internals.compiler.func_compiler import compile_global_func_def
from guppylang_internals.definition.common import (
    CheckableDef,
    MonomorphizableDef,
    MonomorphizedDef,
    ParsableDef,
    UnknownSourceError,
)
from guppylang_internals.definition.value import (
    CallableDef,
    CallReturnWires,
    CompiledCallableDef,
    CompiledHugrNodeDef,
)
from guppylang_internals.error import GuppyError
from guppylang_internals.nodes import GlobalCall
from guppylang_internals.span import SourceMap
from guppylang_internals.tys.subst import Inst, Subst
from guppylang_internals.tys.ty import FunctionType, Type, type_to_row

if TYPE_CHECKING:
    from guppylang_internals.tys.param import Parameter

PyFunc = Callable[..., Any]


@dataclass(frozen=True)
class RawFunctionDef(ParsableDef):
    """A raw function definition provided by the user.

    The raw definition stores exactly what the user has written (i.e. the AST), without
    any additional checking or parsing. Furthermore, we store the values of the Python
    variables in scope at the point of definition.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        python_func: The Python function to be defined.
    """

    python_func: PyFunc

    description: str = field(default="function", init=False)

    def parse(self, globals: Globals, sources: SourceMap) -> "ParsedFunctionDef":
        """Parses and checks the user-provided signature of the function."""
        func_ast, docstring = parse_py_func(self.python_func, sources)
        ty = check_signature(func_ast, globals)
        return ParsedFunctionDef(self.id, self.name, func_ast, ty, docstring)


@dataclass(frozen=True)
class ParsedFunctionDef(CheckableDef, CallableDef):
    """A function definition with parsed and checked signature.

    In particular, this means that we have determined a type for the function and are
    ready to check the function body.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        ty: The type of the function.
        python_scope: The Python scope where the function was defined.
        docstring: The docstring of the function.
    """

    defined_at: ast.FunctionDef
    ty: FunctionType
    docstring: str | None

    description: str = field(default="function", init=False)

    def check(self, globals: Globals) -> "CheckedFunctionDef":
        """Type checks the body of the function."""
        # Add python variable scope to the globals
        cfg = check_global_func_def(self.defined_at, self.ty, globals)
        return CheckedFunctionDef(
            self.id,
            self.name,
            self.defined_at,
            self.ty,
            self.docstring,
            cfg,
        )

    def check_call(
        self, args: list[ast.expr], ty: Type, node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Subst]:
        """Checks the return type of a function call against a given type."""
        # Use default implementation from the expression checker
        args, subst, inst = check_call(self.ty, args, ty, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return node, subst

    def synthesize_call(
        self, args: list[ast.expr], node: AstNode, ctx: Context
    ) -> tuple[ast.expr, Type]:
        """Synthesizes the return type of a function call."""
        # Use default implementation from the expression checker
        args, ty, inst = synthesize_call(self.ty, args, node, ctx)
        node = with_loc(node, GlobalCall(def_id=self.id, args=args, type_args=inst))
        return with_type(ty, node), ty


@dataclass(frozen=True)
class CheckedFunctionDef(ParsedFunctionDef, MonomorphizableDef):
    """Type checked version of a user-defined function that is ready to be compiled.

    In particular, this means that we have a constructed and type checked a control-flow
    graph for the function body.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        ty: The type of the function.
        python_scope: The Python scope where the function was defined.
        docstring: The docstring of the function.
        cfg: The type- and linearity-checked CFG for the function body.
    """

    cfg: CheckedCFG[Place]

    @property
    def params(self) -> "Sequence[Parameter]":
        """Generic parameters of this function."""
        return self.ty.params

    def monomorphize(
        self,
        module: DefinitionBuilder[OpVar],
        mono_args: "PartiallyMonomorphizedArgs",
        ctx: "CompilerContext",
    ) -> "CompiledFunctionDef":
        """Adds a Hugr `FuncDefn` node for the (partially) monomorphized function to the
        Hugr.

        Note that we don't compile the function body at this point since we don't have
        access to the other compiled functions yet. The body is compiled later in
        `CompiledFunctionDef.compile_inner()`.
        """
        mono_ty = self.ty.instantiate_partial(mono_args)
        hugr_ty = mono_ty.to_hugr_poly(ctx)
        func_def = module.module_root_builder().define_function(
            self.name, hugr_ty.body.input, hugr_ty.body.output, hugr_ty.params
        )
        return CompiledFunctionDef(
            self.id,
            self.name,
            self.defined_at,
            mono_args,
            mono_ty,
            self.docstring,
            self.cfg,
            func_def,
        )


@dataclass(frozen=True)
class CompiledFunctionDef(
    CheckedFunctionDef, CompiledCallableDef, MonomorphizedDef, CompiledHugrNodeDef
):
    """A function definition with a corresponding Hugr node.

    Args:
        id: The unique definition identifier.
        name: The name of the function.
        defined_at: The AST node where the function was defined.
        mono_args: Partial monomorphization of the generic type parameters.
        ty: The type of the function after partial monomorphization.
        python_scope: The Python scope where the function was defined.
        docstring: The docstring of the function.
        cfg: The type- and linearity-checked CFG for the function body.
        func_def: The Hugr function definition.
    """

    func_def: hf.Function

    @property
    def hugr_node(self) -> Node:
        """The Hugr node this definition was compiled into."""
        return self.func_def.parent_node

    def load_with_args(
        self,
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> Wire:
        """Loads the function as a value into a local Hugr dataflow graph."""
        return load_with_args(type_args, dfg, self.ty, self.func_def)

    def compile_call(
        self,
        args: list[Wire],
        type_args: Inst,
        dfg: DFContainer,
        ctx: CompilerContext,
        node: AstNode,
    ) -> CallReturnWires:
        """Compiles a call to the function."""
        return compile_call(args, type_args, dfg, self.ty, self.func_def)

    def compile_inner(self, globals: CompilerContext) -> None:
        """Compiles the body of the function."""
        compile_global_func_def(self, self.func_def, globals)


def load_with_args(
    type_args: Inst,
    dfg: DFContainer,
    ty: FunctionType,
    func: ToNode,
) -> Wire:
    """Loads the function as a value into a local Hugr dataflow graph."""
    func_ty: ht.FunctionType = ty.instantiate(type_args).to_hugr(dfg.ctx)
    type_args = [ta.to_hugr(dfg.ctx) for ta in type_args]
    return dfg.builder.load_function(func, func_ty, type_args)


def compile_call(
    args: list[Wire],
    type_args: Inst,  # Non-monomorphized type args only
    dfg: DFContainer,
    ty: FunctionType,
    func: ToNode,
) -> CallReturnWires:
    """Compiles a call to the function."""
    func_ty: ht.FunctionType = ty.instantiate(type_args).to_hugr(dfg.ctx)
    type_args = [arg.to_hugr(dfg.ctx) for arg in type_args]
    num_returns = len(type_to_row(ty.output))
    call = dfg.builder.call(func, *args, instantiation=func_ty, type_args=type_args)
    return CallReturnWires(
        regular_returns=list(call[:num_returns]),
        inout_returns=list(call[num_returns:]),
    )


def parse_py_func(f: PyFunc, sources: SourceMap) -> tuple[ast.FunctionDef, str | None]:
    source_lines, line_offset = inspect.getsourcelines(f)
    source, func_ast, line_offset = parse_source(source_lines, line_offset)
    file = inspect.getsourcefile(f)
    if file is None:
        raise GuppyError(UnknownSourceError(None, f))
    sources.add_file(file)
    annotate_location(func_ast, source, file, line_offset)
    if not isinstance(func_ast, ast.FunctionDef):
        raise GuppyError(ExpectedError(func_ast, "a function definition"))
    return parse_function_with_docstring(func_ast)


def parse_source(source_lines: list[str], line_offset: int) -> tuple[str, ast.AST, int]:
    """Parses a list of source lines into an AST object.

    Also takes care of correctly parsing source that is indented.

    Returns the full source, the parsed AST node, and a potentially updated line number
    offset.
    """
    source = "".join(source_lines)  # Lines already have trailing \n's
    if source_lines[0][0].isspace():
        # This means the function is indented, so we cannot parse it straight away.
        # Running `textwrap.dedent` would mess up the column number in spans. Instead,
        # we'll just wrap the source into a dummy class definition so the indent becomes
        # valid
        cls_node = ast.parse("class _:\n" + source).body[0]
        assert isinstance(cls_node, ast.ClassDef)
        node = cls_node.body[0]
        line_offset -= 1
    else:
        node = ast.parse(source).body[0]
    return source, node, line_offset



from guppylang.decorator import guppy
from guppylang_internals.decorator import hugr_op
from guppylang_internals.std._internal.util import int_op
from guppylang_internals.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NumericType,
)
from guppylang_internals.error import GuppyTypeError

# Define custom signature for iadd operation
ty = FunctionType(
    params=[],
    inputs=[
        FuncInput(NumericType(kind=NumericType.Kind.Int), InputFlags.NoFlags),
        FuncInput(NumericType(kind=NumericType.Kind.Int), InputFlags.NoFlags),
    ],
    output=NumericType(kind=NumericType.Kind.Int),
)


# Create hugr_op using custom signature
@hugr_op(int_op("iadd"), signature=ty)
def fake_iadd(*args) -> None: ...


def test_custom_signature_success(validate):
    @guppy
    def test_func() -> None:
        fake_iadd(1, 1)

    validate(test_func.compile())


def test_custom_signature_fail(validate):
    @guppy
    def test_func_fail() -> None:
        fake_iadd(1.0, 1.0)

    try:
        validate(test_func_fail.compile())
    except GuppyTypeError as e:
        assert True



import pathlib
import pytest
from guppylang import guppy
from guppylang.std.quantum import qubit

from tests.error.util import run_error_test

path = pathlib.Path(__file__).parent.resolve() / "overload_errors"
files = [
    x
    for x in path.iterdir()
    if x.is_file() and x.suffix == ".py" and x.name != "__init__.py"
]

# Turn paths into strings, otherwise pytest doesn't display the names
files = [str(f) for f in files]


@pytest.mark.parametrize("file", files)
def test_overload_errors(file, capsys, snapshot):
    run_error_test(file, capsys, snapshot)


def test_not_enough_overloads():
    @guppy.declare
    def foo() -> None: ...

    with pytest.raises(ValueError, match="Overload requires at least two functions"):
        @guppy.overload()
        def overloaded(): ...

    with pytest.raises(ValueError, match="Overload requires at least two functions"):
        @guppy.overload
        def overloaded(): ...

    with pytest.raises(ValueError, match="Overload requires at least two functions"):
        @guppy.overload(foo)
        def overloaded(): ...


def test_non_guppy_overload():
    @guppy.declare
    def foo() -> None: ...

    with pytest.raises(TypeError, match="Not a Guppy definition: 42"):
        @guppy.overload(foo, 42)
        def overloaded(): ...


def test_non_function_overload():
    @guppy.declare
    def foo() -> None: ...

    with pytest.raises(
        TypeError, match="Not a Guppy function definition: type `qubit`"
    ):
        @guppy.overload(foo, qubit)
        def overloaded(): ...



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


blah = owned


@guppy.declare
def foo(x: int @blah) -> qubit: ...


foo.compile()



from collections.abc import Callable

from guppylang.decorator import guppy
from guppylang.std.builtins import array, comptime

from hugr import ops
from hugr.std.int import IntVal


def test_flat(validate):
    @guppy.comptime
    def foo() -> int:
        x = 0
        for i in range(10):
            x += i
        return x

    hugr = foo.compile().modules[0]
    assert hugr.num_nodes() == 6
    [const] = [data.op for _, data in hugr.nodes() if isinstance(data.op, ops.Const)]
    assert isinstance(const.val, IntVal)
    assert const.val.v == sum(i for i in range(10))
    validate(hugr)


def test_inputs(validate):
    @guppy.comptime
    def foo(x: int, y: float) -> tuple[int, float]:
        return x, y

    validate(foo.compile())


def test_recursion(validate):
    @guppy.comptime
    def foo(x: int) -> int:
        # `foo` doesn't terminate but the compiler should!
        return foo(x)

    validate(foo.compile())


def test_calls(validate):
    @guppy.comptime
    def comptime1(x: int) -> int:
        return regular1(x)

    @guppy
    def regular1(x: int) -> int:
        return comptime2(x)

    @guppy.comptime
    def comptime2(x: int) -> int:
        return regular2(x)

    @guppy
    def regular2(x: int) -> int:
        return comptime1(x)

    validate(regular2.compile())


def test_load_func(validate):
    @guppy.declare
    def foo(x: int) -> int: ...

    @guppy.comptime
    def test() -> Callable[[int], int]:
        return foo

    validate(test.compile())


def test_inner_scope(validate):
    def make(n: int):
        # Test that `n` is scope even if it is only defined in this function scope
        # instead of globally:

        @guppy.comptime
        def foo() -> int:
            return n

        @guppy.comptime
        def bar(xs: array[int, comptime(n)]) -> None:
            pass

        return foo, bar

    foo, bar = make(42)
    validate(foo.compile())
    validate(bar.compile())



from hugr import Wire, ops

from guppylang_internals.definition.custom import CustomInoutCallCompiler
from guppylang_internals.definition.value import CallReturnWires


class WithOwnedCompiler(CustomInoutCallCompiler):
    """Compiler for the `with_owned` function."""

    def compile_with_inouts(self, args: list[Wire]) -> CallReturnWires:
        [val, func] = args
        [out, val] = self.builder.add_op(ops.CallIndirect(), func, val)
        return CallReturnWires(regular_returns=[out], inout_returns=[val])



from guppylang import guppy
from guppylang_internals.decorator import extend_type
from tests.integration.modules.mod_a import f, MyType


@guppy.declare
def g() -> MyType: ...


@guppy
def h(x: int) -> int:
    return f(x)


# Extend type defined in module A
@extend_type(MyType)
class _:
    @guppy
    def __int__(self: "MyType") -> int:
        return 0



import ast
import copy
import itertools
from collections.abc import Iterable, Iterator
from dataclasses import dataclass, field, replace
from functools import cache, cached_property
from types import FrameType
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    NamedTuple,
    TypeAlias,
    TypeVar,
    cast,
    overload,
)

from typing_extensions import assert_never

from guppylang_internals.ast_util import AstNode, name_nodes_in_ast
from guppylang_internals.cfg.bb import VId
from guppylang_internals.definition.common import (
    DefId,
    Definition,
    ParsedDef,
)
from guppylang_internals.definition.ty import TypeDef
from guppylang_internals.definition.value import CallableDef
from guppylang_internals.engine import BUILTIN_DEFS, DEF_STORE, ENGINE
from guppylang_internals.error import InternalGuppyError
from guppylang_internals.tys.builtin import (
    callable_type_def,
    float_type_def,
    int_type_def,
    nat_type_def,
    none_type_def,
    tuple_type_def,
)
from guppylang_internals.tys.param import Parameter
from guppylang_internals.tys.ty import (
    BoundTypeVar,
    ExistentialTypeVar,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
    OpaqueType,
    StructType,
    SumType,
    TupleType,
    Type,
)

if TYPE_CHECKING:
    from guppylang_internals.definition.struct import StructField


#: A "place" is a description for a storage location of a local value that users
#: can refer to in their program.
#:
#: Roughly, these are values that can be lowered to a static wire within the Hugr
#: representation. The most basic example of a place is a single local variable. Beyond
#: that, we also treat some projections of local variables (e.g. nested struct field
#: accesses) as places.
#:
#: All places are equipped with a unique id, a type and an optional definition AST
#: location. During linearity checking, they are tracked separately.
Place: TypeAlias = "Variable | FieldAccess | SubscriptAccess | TupleAccess"

#: Unique identifier for a `Place`.
PlaceId: TypeAlias = (
    "Variable.Id | FieldAccess.Id | SubscriptAccess.Id | TupleAccess.Id"
)


@dataclass(frozen=True)
class Variable:
    """A place identifying a local variable."""

    name: str
    ty: Type
    defined_at: AstNode | None
    flags: InputFlags = InputFlags.NoFlags

    # Remember if a variable is a function input. This way, we can e.g. suggest to add a
    # `@comptime` annotation to users when appropriate
    is_func_input: bool = field(default=False, kw_only=True)

    @dataclass(frozen=True)
    class Id:
        """Identifier for variable places."""

        name: str

    @cached_property
    def id(self) -> "Variable.Id":
        """The unique `PlaceId` identifier for this place."""
        return Variable.Id(self.name)

    @cached_property
    def root(self) -> "Variable":
        """The root variable of this place."""
        return self

    @property
    def describe(self) -> str:
        """A human-readable description of this place for error messages."""
        return f"Variable `{self}`"

    def __str__(self) -> str:
        """String representation of this place."""
        return self.name

    def replace_defined_at(self, node: AstNode | None) -> "Variable":
        """Returns a new `Variable` instance with an updated definition location."""
        return replace(self, defined_at=node)


@dataclass(frozen=True, kw_only=True)
class ComptimeVariable(Variable):
    """A place identifying a variable that is passed from a compile-time context."""

    # Store the static value of the variable.
    static_value: Any


@dataclass(frozen=True)
class FieldAccess:
    """A place identifying a field access on a local struct."""

    parent: Place
    field: "StructField"
    exact_defined_at: AstNode | None

    @dataclass(frozen=True)
    class Id:
        """Identifier for field places."""

        parent: PlaceId
        field: str

    def __post_init__(self) -> None:
        # Check that the field access is consistent
        assert self.struct_ty.field_dict[self.field.name] == self.field

    @cached_property
    def id(self) -> "FieldAccess.Id":
        """The unique `PlaceId` identifier for this place."""
        return FieldAccess.Id(self.parent.id, self.field.name)

    @cached_property
    def root(self) -> "Variable":
        """The root variable of this place."""
        return self.parent.root

    @property
    def ty(self) -> Type:
        """The type of this place."""
        return self.field.ty

    @cached_property
    def struct_ty(self) -> StructType:
        """The type of the struct whose field is accessed."""
        assert isinstance(self.parent.ty, StructType)
        return self.parent.ty

    @cached_property
    def defined_at(self) -> AstNode | None:
        """Optional location where this place was last assigned to."""
        return self.exact_defined_at or self.parent.defined_at

    @cached_property
    def describe(self) -> str:
        """A human-readable description of this place for error messages."""
        return f"Field `{self}`"

    def __str__(self) -> str:
        """String representation of this place."""
        return f"{self.parent}.{self.field.name}"

    def replace_defined_at(self, node: AstNode | None) -> "FieldAccess":
        """Returns a new `FieldAccess` instance with an updated definition location."""
        return replace(self, exact_defined_at=node)


class SetitemCall(NamedTuple):
    """
    Represents a `__setitem__` call for assigning values to array elements.

    Holds the expression corresponding to the `__setitem__` call and the variable
    holding the value to be written to the subscript.
    """

    #: Expression corresponding to the `__setitem__` call.
    call: ast.expr

    #: Variable holding the value that should be written to the subscript.
    #: This variable *must be* assigned before compiling the call!
    value_var: Variable


@dataclass(frozen=True)
class SubscriptAccess:
    """A place identifying a subscript `place[item]` access."""

    parent: Place
    item: Variable
    ty: Type
    item_expr: ast.expr
    getitem_call: ast.expr | None = None
    setitem_call: SetitemCall | None = None

    @dataclass(frozen=True)
    class Id:
        """Identifier for subscript places."""

        parent: PlaceId
        item: Variable.Id

    @cached_property
    def id(self) -> "SubscriptAccess.Id":
        """The unique `PlaceId` identifier for this place."""
        return SubscriptAccess.Id(self.parent.id, self.item.id)

    @cached_property
    def defined_at(self) -> AstNode | None:
        """Optional location where this place was last assigned to."""
        return self.parent.defined_at

    @cached_property
    def root(self) -> "Variable":
        """The root variable of this place."""
        return self.parent.root

    @property
    def describe(self) -> str:
        """A human-readable description of this place for error messages."""
        return f"Subscript `{self}`"

    def __str__(self) -> str:
        """String representation of this place."""
        return f"{self.parent}[...]"


def contains_subscript(place: Place) -> SubscriptAccess | None:
    """Checks if a place contains a subscript access and returns the rightmost one."""
    while not isinstance(place, Variable):
        if isinstance(place, SubscriptAccess):
            return place
        place = place.parent
    return None


@dataclass(frozen=True)
class TupleAccess:
    """A place identifying an index access on a tuple."""

    parent: Place
    elem_ty: Type
    index: int
    exact_defined_at: AstNode | None

    @dataclass(frozen=True)
    class Id:
        """Identifier for tuple places."""

        parent: PlaceId
        index: int

    @cached_property
    def id(self) -> "TupleAccess.Id":
        """The unique `PlaceId` identifier for this place."""
        return TupleAccess.Id(self.parent.id, self.index)

    @cached_property
    def root(self) -> "Variable":
        """The root variable of this place."""
        return self.parent.root

    @property
    def ty(self) -> Type:
        """The type of this place."""
        return self.elem_ty

    @cached_property
    def defined_at(self) -> AstNode | None:
        """Optional location where this place was last assigned to."""
        return self.exact_defined_at or self.parent.defined_at

    @cached_property
    def describe(self) -> str:
        """A human-readable description of this place for error messages."""
        return f"Tuple index `{self}`"

    def __str__(self) -> str:
        """String representation of this place."""
        return f"{self.parent}[{self.index}]"

    def replace_defined_at(self, node: AstNode | None) -> "TupleAccess":
        """Returns a new `TupleAccess` instance with an updated definition location."""
        return replace(self, exact_defined_at=node)


@dataclass(frozen=True)
class PythonObject:
    """Wrapper around an arbitrary Python object.

    Used to distinguish between Guppy definitions and other values when looking up
    object from the enclosing Python scope by name.
    """

    obj: Any


class Globals:
    """Wrapper around the `DEF_STORE` that allows looking-up of definitions by name
    based on which objects are in scope in a stack frame.

    Additionally, keeps track of which definitions in the store have been used.
    """

    f_locals: dict[str, Any]
    f_globals: dict[str, Any]
    f_builtins: dict[str, Any]

    def __init__(self, frame: FrameType | None) -> None:
        if frame is not None:
            self.f_locals = frame.f_locals
            self.f_globals = frame.f_globals
            self.f_builtins = frame.f_builtins
        else:
            self.f_locals = {}
            self.f_globals = {}
            self.f_builtins = {}

    @staticmethod
    @cache
    def builtin_defs() -> dict[str, Definition]:
        import guppylang.std.builtins
        from guppylang.defs import GuppyDefinition

        return BUILTIN_DEFS | {
            name: val.wrapped
            for name, val in guppylang.std.builtins.__dict__.items()
            if isinstance(val, GuppyDefinition)
        }

    def get_instance_func(self, ty: Type | TypeDef, name: str) -> CallableDef | None:
        """Looks up an instance function with a given name for a type.

        Returns `None` if the name doesn't exist or isn't a function.
        """
        type_defn: TypeDef
        match ty:
            case TypeDef() as type_defn:
                pass
            case BoundTypeVar() | ExistentialTypeVar() | SumType():
                return None
            case NumericType(kind):
                match kind:
                    case NumericType.Kind.Nat:
                        type_defn = nat_type_def
                    case NumericType.Kind.Int:
                        type_defn = int_type_def
                    case NumericType.Kind.Float:
                        type_defn = float_type_def
                    case kind:
                        return assert_never(kind)
            case FunctionType():
                type_defn = callable_type_def
            case OpaqueType() as ty:
                type_defn = ty.defn
            case StructType() as ty:
                type_defn = ty.defn
            case TupleType():
                type_defn = tuple_type_def
            case NoneType():
                type_defn = none_type_def
            case _:
                return assert_never(ty)

        type_defn = cast(TypeDef, ENGINE.get_checked(type_defn.id))
        if type_defn.id in DEF_STORE.impls and name in DEF_STORE.impls[type_defn.id]:
            def_id = DEF_STORE.impls[type_defn.id][name]
            defn = ENGINE.get_parsed(def_id)
            if isinstance(defn, CallableDef):
                return defn
        return None

    def __contains__(self, item: DefId | str) -> bool:
        match item:
            case DefId() as def_id:
                return def_id in DEF_STORE.raw_defs
            case str(x):
                return (
                    x in self.builtin_defs()
                    or x in self.f_locals
                    or x in self.f_globals
                )
            case x:
                return assert_never(x)

    @overload
    def __getitem__(self, item: DefId) -> ParsedDef: ...

    @overload
    def __getitem__(self, item: str) -> "ParsedDef | PythonObject": ...

    def __getitem__(self, item: DefId | str) -> "ParsedDef | PythonObject":
        from guppylang.defs import GuppyDefinition

        match item:
            case DefId() as def_id:
                return ENGINE.get_parsed(def_id)
            case str(name):
                if name in self.f_locals:
                    val = self.f_locals[name]
                    if isinstance(val, GuppyDefinition):
                        return ENGINE.get_parsed(val.id)
                    # Before falling back to returning the Python object, check if we
                    # have defined the name as a builtin
                    elif name in self.builtin_defs():
                        defn = self.builtin_defs()[name]
                        return ENGINE.get_parsed(defn.id)
                    else:
                        return PythonObject(val)
                elif name in self.f_globals:
                    val = self.f_globals[name]
                    if isinstance(val, GuppyDefinition):
                        return ENGINE.get_parsed(val.id)
                    # Before falling back to returning the Python object, check if we
                    # have defined the name as a builtin
                    elif name in self.builtin_defs():
                        defn = self.builtin_defs()[name]
                        return ENGINE.get_parsed(defn.id)
                    else:
                        return PythonObject(val)
                elif name in self.builtin_defs():
                    defn = self.builtin_defs()[name]
                    return ENGINE.get_parsed(defn.id)
                else:
                    raise InternalGuppyError(f"Cannot find definition `{name}`")
            case x:
                return assert_never(x)


V = TypeVar("V")


@dataclass
class Locals(Generic[VId, V]):
    """Scoped mapping from program variable ids to the corresponding program variable.

    Depending on which checking phase we are in (type checking or linearity checking),
    we use this either as a mapping from strings to `Variable`s or as a mapping from
    `PlaceId`s to `Place`s.
    """

    vars: dict[VId, V]
    parent_scope: "Locals[VId, V] | None" = None

    def __getitem__(self, item: VId) -> V:
        if item not in self.vars and self.parent_scope:
            return self.parent_scope[item]

        return self.vars[item]

    def __setitem__(self, key: VId, value: V) -> None:
        self.vars[key] = value

    def __iter__(self) -> Iterator[VId]:
        parent_iter = iter(self.parent_scope) if self.parent_scope else iter(())
        return itertools.chain(iter(self.vars), parent_iter)

    def __contains__(self, item: VId) -> bool:
        return (item in self.vars) or (
            self.parent_scope is not None and item in self.parent_scope
        )

    def __copy__(self) -> "Locals[VId, V]":
        # Make a copy of the var map so that mutating the copy doesn't
        # mutate our variable mapping
        return Locals(self.vars.copy(), copy.copy(self.parent_scope))

    def keys(self) -> set[VId]:
        parent_keys = self.parent_scope.keys() if self.parent_scope else set()
        return parent_keys | self.vars.keys()

    def values(self) -> Iterable[V]:
        parent_values = (
            iter(self.parent_scope.values()) if self.parent_scope else iter(())
        )
        return itertools.chain(self.vars.values(), parent_values)

    def items(self) -> Iterable[tuple[VId, V]]:
        parent_items = (
            iter(self.parent_scope.items()) if self.parent_scope else iter(())
        )
        return itertools.chain(self.vars.items(), parent_items)


class Context(NamedTuple):
    """The type checking context."""

    globals: Globals
    locals: Locals[str, Variable]
    generic_params: dict[str, Parameter]


class DummyEvalDict(dict[str, Any]):
    """A custom dict that can be passed to `eval` to give better error messages.
    This class is used to implement the `py(...)` expression. If the user tries to
    access a Guppy variable in the Python context, we give an informative error message.
    """

    ctx: Context
    node: ast.expr

    @dataclass
    class GuppyVarUsedError(BaseException):
        """Error that is raised when the user tries to access a Guppy variable."""

        var: str
        node: ast.Name | None

    def __init__(self, ctx: Context, node: ast.expr):
        super().__init__(**(ctx.globals.f_globals | ctx.globals.f_locals))
        self.ctx = ctx
        self.node = node

    def _check_item(self, key: str) -> None:
        # Catch the user trying to access Guppy variables
        if key in self.ctx.locals:
            # Find the name node in the AST where the usage occurs
            n = next((n for n in name_nodes_in_ast(self.node) if n.id == key), None)
            raise self.GuppyVarUsedError(key, n)

    def __getitem__(self, key: str) -> Any:
        self._check_item(key)
        return super().__getitem__(key)

    def __delitem__(self, key: str) -> None:
        self._check_item(key)
        super().__delitem__(key)

    def __contains__(self, key: object) -> bool:
        if isinstance(key, str) and key in self.ctx.locals:
            return True
        return super().__contains__(key)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, ParamSpec, TypeVar

from hugr import ops
from hugr import tys as ht

from guppylang_internals.compiler.core import (
    CompilerContext,
    GlobalConstId,
)
from guppylang_internals.definition.common import DefId
from guppylang_internals.definition.custom import (
    CustomCallChecker,
    CustomFunctionDef,
    CustomInoutCallCompiler,
    DefaultCallChecker,
    NotImplementedCallCompiler,
    OpCompiler,
    RawCustomFunctionDef,
)
from guppylang_internals.definition.ty import OpaqueTypeDef, TypeDef
from guppylang_internals.definition.wasm import RawWasmFunctionDef
from guppylang_internals.dummy_decorator import _dummy_custom_decorator, sphinx_running
from guppylang_internals.engine import DEF_STORE
from guppylang_internals.std._internal.checker import WasmCallChecker
from guppylang_internals.std._internal.compiler.wasm import (
    WasmModuleCallCompiler,
    WasmModuleDiscardCompiler,
    WasmModuleInitCompiler,
)
from guppylang_internals.tys.builtin import (
    WasmModuleTypeDef,
)
from guppylang_internals.tys.ty import (
    FuncInput,
    FunctionType,
    InputFlags,
    NoneType,
    NumericType,
)

if TYPE_CHECKING:
    import builtins
    from collections.abc import Callable, Sequence
    from types import FrameType

    from guppylang.defs import GuppyDefinition, GuppyFunctionDefinition
    from guppylang_internals.tys.arg import Argument
    from guppylang_internals.tys.param import Parameter
    from guppylang_internals.tys.subst import Inst

T = TypeVar("T")
P = ParamSpec("P")


def get_calling_frame() -> FrameType:
    """Finds the first frame that called this function outside the compiler modules."""
    frame = inspect.currentframe()
    while frame:
        module = inspect.getmodule(frame)
        if module is None:
            return frame
        if module.__file__ != __file__:
            return frame
        frame = frame.f_back
    raise RuntimeError("Couldn't obtain stack frame for definition")


def custom_function(
    compiler: CustomInoutCallCompiler | None = None,
    checker: CustomCallChecker | None = None,
    higher_order_value: bool = True,
    name: str = "",
    signature: FunctionType | None = None,
) -> Callable[[Callable[P, T]], GuppyFunctionDefinition[P, T]]:
    """Decorator to add custom typing or compilation behaviour to function decls.

    Optionally, usage of the function as a higher-order value can be disabled. In
    that case, the function signature can be omitted if a custom call compiler is
    provided.
    """
    from guppylang.defs import GuppyFunctionDefinition

    def dec(f: Callable[P, T]) -> GuppyFunctionDefinition[P, T]:
        call_checker = checker or DefaultCallChecker()
        func = RawCustomFunctionDef(
            DefId.fresh(),
            name or f.__name__,
            None,
            f,
            call_checker,
            compiler or NotImplementedCallCompiler(),
            higher_order_value,
            signature,
        )
        DEF_STORE.register_def(func, get_calling_frame())
        return GuppyFunctionDefinition(func)

    return dec


def hugr_op(
    op: Callable[[ht.FunctionType, Inst, CompilerContext], ops.DataflowOp],
    checker: CustomCallChecker | None = None,
    higher_order_value: bool = True,
    name: str = "",
    signature: FunctionType | None = None,
) -> Callable[[Callable[P, T]], GuppyFunctionDefinition[P, T]]:
    """Decorator to annotate function declarations as HUGR ops.

    Args:
        op: A function that takes an instantiation of the type arguments as well as
            the inferred input and output types and returns a concrete HUGR op.
        checker: The custom call checker.
        higher_order_value: Whether the function may be used as a higher-order
            value.
        name: The name of the function.
    """
    return custom_function(OpCompiler(op), checker, higher_order_value, name, signature)


def extend_type(defn: TypeDef) -> Callable[[type], type]:
    """Decorator to add new instance functions to a type."""
    from guppylang.defs import GuppyDefinition

    def dec(c: type) -> type:
        for val in c.__dict__.values():
            if isinstance(val, GuppyDefinition):
                DEF_STORE.register_impl(defn.id, val.wrapped.name, val.id)
        return c

    return dec


def custom_type(
    hugr_ty: ht.Type | Callable[[Sequence[Argument], CompilerContext], ht.Type],
    name: str = "",
    copyable: bool = True,
    droppable: bool = True,
    bound: ht.TypeBound | None = None,
    params: Sequence[Parameter] | None = None,
) -> Callable[[type[T]], type[T]]:
    """Decorator to annotate a class definitions as Guppy types.

    Requires the static Hugr translation of the type. Additionally, the type can be
    marked as linear. All `@guppy` annotated functions on the class are turned into
    instance functions.

    For non-generic types, the Hugr representation can be passed as a static value.
    For generic types, a callable may be passed that takes the type arguments of a
    concrete instantiation.
    """
    from guppylang.defs import GuppyDefinition

    mk_hugr_ty = (
        (lambda args, ctx: hugr_ty) if isinstance(hugr_ty, ht.Type) else hugr_ty
    )

    def dec(c: type[T]) -> type[T]:
        defn = OpaqueTypeDef(
            DefId.fresh(),
            name or c.__name__,
            None,
            params or [],
            not copyable,
            not droppable,
            mk_hugr_ty,
            bound,
        )
        DEF_STORE.register_def(defn, get_calling_frame())
        for val in c.__dict__.values():
            if isinstance(val, GuppyDefinition):
                DEF_STORE.register_impl(defn.id, val.wrapped.name, val.id)
        # We're pretending to return the class unchanged, but in fact we return
        # a `GuppyDefinition` that handles the comptime logic
        return GuppyDefinition(defn)  # type: ignore[return-value]

    return dec


def wasm_module(
    filename: str, filehash: int
) -> Callable[[builtins.type[T]], GuppyDefinition]:
    from guppylang.defs import GuppyDefinition

    def dec(cls: builtins.type[T]) -> GuppyDefinition:
        # N.B. Only one module per file and vice-versa
        wasm_module = WasmModuleTypeDef(
            DefId.fresh(),
            cls.__name__,
            None,
            filename,
            filehash,
        )

        wasm_module_ty = wasm_module.check_instantiate([], None)

        DEF_STORE.register_def(wasm_module, get_calling_frame())
        for val in cls.__dict__.values():
            if isinstance(val, GuppyDefinition):
                DEF_STORE.register_impl(wasm_module.id, val.wrapped.name, val.id)
        # Add a constructor to the class
        call_method = CustomFunctionDef(
            DefId.fresh(),
            "__new__",
            None,
            FunctionType(
                [FuncInput(NumericType(NumericType.Kind.Nat), flags=InputFlags.Owned)],
                wasm_module_ty,
            ),
            DefaultCallChecker(),
            WasmModuleInitCompiler(),
            True,
            GlobalConstId.fresh(f"{cls.__name__}.__new__"),
            True,
        )
        discard = CustomFunctionDef(
            DefId.fresh(),
            "discard",
            None,
            FunctionType([FuncInput(wasm_module_ty, InputFlags.Owned)], NoneType()),
            DefaultCallChecker(),
            WasmModuleDiscardCompiler(),
            False,
            GlobalConstId.fresh(f"{cls.__name__}.__discard__"),
            True,
        )
        DEF_STORE.register_def(call_method, get_calling_frame())
        DEF_STORE.register_impl(wasm_module.id, "__new__", call_method.id)
        DEF_STORE.register_def(discard, get_calling_frame())
        DEF_STORE.register_impl(wasm_module.id, "discard", discard.id)

        return GuppyDefinition(wasm_module)

    return dec


def wasm(f: Callable[P, T]) -> GuppyFunctionDefinition[P, T]:
    from guppylang.defs import GuppyFunctionDefinition

    func = RawWasmFunctionDef(
        DefId.fresh(),
        f.__name__,
        None,
        f,
        WasmCallChecker(),
        WasmModuleCallCompiler(f.__name__),
        True,
        signature=None,
    )
    DEF_STORE.register_def(func, get_calling_frame())
    return GuppyFunctionDefinition(func)


# Override decorators with dummy versions if we're running a sphinx build
if not TYPE_CHECKING and sphinx_running():
    custom_function = _dummy_custom_decorator
    hugr_op = _dummy_custom_decorator
    extend_type = _dummy_custom_decorator
    custom_type = _dummy_custom_decorator
    wasm_module = _dummy_custom_decorator
    wasm = _dummy_custom_decorator()



from guppylang import guppy
from guppylang.std.builtins import nat, comptime, array


@guppy.declare
def foo(n: nat @comptime, xs: "array[int, n]") -> None: ...


@guppy
def main() -> None:
    foo(3, array(1, 2))


main.compile()



from guppylang import guppy
from guppylang_internals.decorator import wasm, wasm_module
from guppylang.module import GuppyModule

@wasm_module("", 0)
class Foo:
    @wasm
    def foo(self: "Foo") -> bool: ...

@guppy
def main() -> bool:
    f = Foo(0)
    return f.foo()

main.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy
def foo(qs: list[qubit] @owned) -> list[int]:
    return [42 for q in qs]


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit, cx


@guppy.comptime
def test(qs: array[qubit, 10]) -> None:
    cx(qs[0], qs[0])


test.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit


@guppy.struct
class MyStruct:
    q: qubit


@guppy.declare
def use(q: qubit @owned) -> None: ...


@guppy
def test(s: MyStruct) -> None:
    use(s.q)


test.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import owned
from guppylang.std.quantum import qubit
from guppylang.std.quantum_functional import h


@guppy
def foo(q: qubit @owned) -> None:
    h(q)


foo.compile()



import functools
from collections.abc import Sequence
from typing import Any

from guppylang_internals.error import InternalGuppyError
from guppylang_internals.tys.arg import Argument, ConstArg, TypeArg
from guppylang_internals.tys.common import Transformer, Visitor
from guppylang_internals.tys.const import (
    BoundConstVar,
    Const,
    ConstBase,
    ExistentialConstVar,
)
from guppylang_internals.tys.ty import (
    BoundTypeVar,
    ExistentialTypeVar,
    FunctionType,
    Type,
    TypeBase,
)
from guppylang_internals.tys.var import BoundVar, ExistentialVar

Subst = dict[ExistentialVar, Type | Const]
Inst = Sequence[Argument]
PartialInst = Sequence["Argument | None"]


class Substituter(Transformer):
    """Type transformer that applies a substitution of existential variables."""

    def __init__(self, subst: Subst) -> None:
        self.subst = subst

    @functools.singledispatchmethod
    def transform(self, ty: Any) -> Any | None:  # type: ignore[override]
        return None

    @transform.register
    def _transform_ExistentialTypeVar(self, ty: ExistentialTypeVar) -> Type | None:
        s = self.subst.get(ty, None)
        assert not isinstance(s, ConstBase)
        return s

    @transform.register
    def _transform_ExistentialConstVar(self, c: ExistentialConstVar) -> Const | None:
        s = self.subst.get(c, None)
        assert not isinstance(s, TypeBase)
        return s


class Instantiator(Transformer):
    """Type transformer that instantiates bound variables."""

    def __init__(self, inst: Inst) -> None:
        self.inst = inst

    @functools.singledispatchmethod
    def transform(self, ty: Any) -> Any | None:  # type: ignore[override]
        return None

    @transform.register
    def _transform_BoundTypeVar(self, ty: BoundTypeVar) -> Type | None:
        # Instantiate if type for the index is available
        if ty.idx < len(self.inst):
            arg = self.inst[ty.idx]
            assert isinstance(arg, TypeArg)
            return arg.ty

        # Otherwise, lower the de Bruijn index
        return BoundTypeVar(
            ty.display_name, ty.idx - len(self.inst), ty.copyable, ty.droppable
        )

    @transform.register
    def _transform_BoundConstVar(self, c: BoundConstVar) -> Const | None:
        # Instantiate if const value for the index is available
        if c.idx < len(self.inst):
            arg = self.inst[c.idx]
            assert isinstance(arg, ConstArg)
            return arg.const

        # Otherwise, lower the de Bruijn index
        return BoundConstVar(c.ty, c.display_name, c.idx - len(self.inst))

    @transform.register
    def _transform_FunctionType(self, ty: FunctionType) -> Type | None:
        if ty.parametrized:
            raise InternalGuppyError("Tried to instantiate under binder")
        return None


class BoundVarFinder(Visitor):
    """Type visitor that looks for occurrences of bound variables."""

    bound_vars: set[BoundVar]

    def __init__(self) -> None:
        self.bound_vars = set()

    @functools.singledispatchmethod
    def visit(self, ty: Any) -> bool:  # type: ignore[override]
        return False

    @visit.register
    def _transform_BoundTypeVar(self, ty: BoundTypeVar) -> bool:
        self.bound_vars.add(ty)
        return False

    @visit.register
    def _transform_BoundConstVar(self, c: BoundConstVar) -> bool:
        self.bound_vars.add(c)
        return False



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy.declare
def bar(q: qubit @owned) -> int: ...


@guppy.declare
def baz(q: qubit) -> int: ...


@guppy
def foo(qs: list[qubit] @owned) -> list[int]:
    return [baz(q) for q in qs if bar(q)]


foo.compile()



from guppylang.decorator import guppy
from guppylang.std.builtins import array
from guppylang.std.quantum import qubit


@guppy
def main() -> None:
    qs = array(qubit() for _ in range(10)) 
    qs[0] = qubit()  # Leaks the old qubit at index 0


main.compile()


from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Generic, Protocol, TypeVar

from hugr import tys as ht

if TYPE_CHECKING:
    from guppylang_internals.tys.const import BoundConstVar
    from guppylang_internals.tys.param import Parameter
    from guppylang_internals.tys.ty import BoundTypeVar

T = TypeVar("T")


class ToHugrContext(Protocol):
    """Protocol for the context capabilities required to translate Guppy types into Hugr
    types.

    Currently, the only required capability is translating bound type and const
    variables into Hugr, taking into account partial monomorphization.
    """

    def type_var_to_hugr(self, var: "BoundTypeVar") -> ht.Type:
        """Compiles a bound Guppy type variable into a Hugr type.

        Should take care of performing partial monomorphization as specified in the
        current context.
        """

    def const_var_to_hugr(self, var: "BoundConstVar") -> ht.TypeArg:
        """Compiles a bound Guppy const variable into a Hugr type argument.

        Should take care of performing partial monomorphization as specified in the
        current context.
        """


@dataclass(frozen=True)
class QuantifiedToHugrContext(ToHugrContext):
    """Concrete implementation of a `ToHugrContext` that should be used when moving
    inside a quantifier.

    This prevents the outer context from touching variables that are bound by a tighter
    quantifier.
    """

    params: Sequence["Parameter"]

    def type_var_to_hugr(self, var: "BoundTypeVar") -> ht.Type:
        return ht.Variable(var.idx, var.hugr_bound)

    def const_var_to_hugr(self, var: "BoundConstVar") -> ht.TypeArg:
        return ht.VariableArg(var.idx, self.params[var.idx].to_hugr(self))


class ToHugr(ABC, Generic[T]):
    """Abstract base class for objects that have a Hugr representation."""

    @abstractmethod
    def to_hugr(self, ctx: ToHugrContext) -> T:
        """Computes the Hugr representation of the object."""


class Visitor(ABC):
    """Abstract base class for a type visitor that transforms types."""

    @abstractmethod
    def visit(self, arg: Any, /) -> bool:
        """This method is called for each visited type.

        Return `False` to continue the recursive descent.
        """


class Transformer(ABC):
    """Abstract base class for a type visitor that transforms types."""

    @abstractmethod
    def transform(self, arg: Any, /) -> Any | None:
        """Transforms the object.

        Return a transformed type or `None` to continue the recursive visit.
        """


class Visitable(ABC):
    """Abstract base class for objects that can be recursively visited."""

    @abstractmethod
    def visit(self, visitor: Visitor, /) -> None:
        """Accepts a visitor on the object.

        Implementors of this method should first invoke the visitor on the object
        itself. If the visitor doesn't handle the object, the visitor should be passed
        to all relevant members of the object.
        """


class Transformable(Visitable, ABC, Generic[T]):
    """Abstract base class for objects that can be recursively transformed."""

    @abstractmethod
    def transform(self, transformer: Transformer, /) -> T:
        """Accepts a transformer on the object.

        Implementors of this method should first invoke the transformer on the object
        itself. If the visitor doesn't handle the object, the visitor should be used to
        transform all relevant members of the object.
        """



import pytest

from hugr import ops, val

from guppylang.decorator import guppy


def test_extern_float(validate):
    ext = guppy._extern("ext", ty="float")

    @guppy
    def main() -> float:
        return ext + ext

    package = main.compile()
    validate(package)

    hg = package.modules[0]
    [c] = [data.op for n, data in hg.nodes() if isinstance(data.op, ops.Const)]
    assert isinstance(c.val, val.Extension)
    assert c.val.val["symbol"] == "ext"


def test_extern_alt_symbol(validate):
    ext = guppy._extern("ext", ty="int", symbol="foo")

    @guppy
    def main() -> int:
        return ext

    package = main.compile()
    validate(package)

    hg = package.modules[0]
    [c] = [data.op for n, data in hg.nodes() if isinstance(data.op, ops.Const)]
    assert isinstance(c.val, val.Extension)
    assert c.val.val["symbol"] == "foo"


def test_extern_tuple(validate):
    ext = guppy._extern("ext", ty="tuple[int, float]")

    @guppy
    def main() -> float:
        x, y = ext
        return x + y

    validate(main.compile())


@pytest.mark.skip("See https://github.com/CQCL/guppylang/issues/827")
def test_extern_conditional_assign(validate):
    x = guppy._extern("x", ty="int")

    @guppy
    def main(b: bool) -> int:
        if b:
            x = 4
        return x

    validate(main.compile())



from typing import Generic

from guppylang.decorator import guppy
from guppylang.std.quantum import qubit

Q = guppy.const_var("Q", "qubit")


@guppy.struct
class Struct(Generic[Q]):
    pass


@guppy
def main(_: Struct[Q]) -> None:
    pass


main.compile()



from guppylang.decorator import guppy
from guppylang.std.quantum import qubit
from guppylang.std.builtins import owned


@guppy
def foo(qs: list[tuple[qubit, bool]] @owned) -> list[qubit]:
    rs: list[qubit] = []
    for q, b in qs:
        rs += [q]
        if b:
            return []
    return rs


foo.compile()
